# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''
其他共用类模块
'''

from __future__ import division
import time
import win32con
import ctypes
import types
import win32api
import win32gui
import win32event
import win32process
import win32service
import winerror
import pywintypes
import six
from PIL import ImageGrab
from win32con import DESKTOP_SWITCHDESKTOP
 
from testbase.util import Timeout
from qt4c.exceptions import TimeoutError
from qt4c.keyboard import Keyboard
from qt4c import wintypes

if six.PY3:
    unicode = str


_DEFAULT_BUFFER_SIZE = 255

SIZEOF = lambda datatype: ctypes.sizeof(datatype)

class Point(object):
    """坐标点类
    """
    def __init__(self, x_y):
        self._pt = (x_y[0], x_y[1])
        
    @property
    def X(self):
        return self._pt[0]
    
    @property
    def Y(self):
        return self._pt[1]
    
    def __eq__(self, pt):
        if self._pt == pt._pt:
            return True
        else:
            return False
        
    @property
    def All(self):
        return self._pt
    
    
class Rectangle(object):
    """矩形区域类
    """
    def __init__(self, left_top_right_bottom):
        self._rect = (left_top_right_bottom[0], left_top_right_bottom[1], left_top_right_bottom[2], left_top_right_bottom[3])
#        if isinstance(self._rect, )
    
    def __str__(self):
        return str(self._rect)
    
    @property
    def All(self):
        return self._rect
    
    @property
    def Bottom(self):
        return self._rect[3]
        
    @property
    def Center(self):
        return Point(((self._rect[0] + self._rect[2]) // 2, (self._rect[1] + self._rect[3]) // 2))
    
    @property
    def Left(self):
        return self._rect[0]
    
    @property
    def Right(self):
        return self._rect[2]
    
    @property
    def Top(self):
        return self._rect[1]
    
    @property
    def Width(self):
        return self.Right - self.Left
    
    @property
    def Height(self):
        return self.Bottom - self.Top
    
    def isInRect(self, rc):
        """判断此区域是否在参数rc的范围内
        
        :type rc: Rectangle
        :param rc: 包含的区域   
        """
        if rc is None or not isinstance(rc, Rectangle):
            raise TypeError("rc应为Rectangle类型，实际为%s" % type(rc))#UtilError("参数类型错误!")
            
        if self.Left >= rc.Left and\
           self.Top >= rc.Top and\
           self.Right <= rc.Right and\
           self.Bottom <= rc.Bottom:
            return True
        else:
            return False
    
    def highLight(self):
        """高亮此区域
        """
        hwnd = win32gui.GetDesktopWindow()
        hDesktopDC = win32gui.GetWindowDC(hwnd)
        oldRop2 = ctypes.windll.gdi32.SetROP2(hDesktopDC, win32con.R2_NOTXORPEN)
        newPen = win32gui.CreatePen(win32con.PS_SOLID, 3, 0)
        oldPen = win32gui.SelectObject(hDesktopDC, newPen)
        #在指示窗口周围显示闪烁矩形
        for i in range(2):
            win32gui.Rectangle(hDesktopDC, self._rect[0], self._rect[1], self._rect[2], self._rect[3])
            time.sleep(0.2)
            win32gui.Rectangle(hDesktopDC, self._rect[0], self._rect[1], self._rect[2], self._rect[3])
            time.sleep(0.3)
        ctypes.windll.gdi32.SetROP2(hDesktopDC, oldRop2)
        win32gui.SelectObject(hDesktopDC, oldPen)
        win32gui.DeleteObject(newPen)
        win32gui.ReleaseDC(hwnd,hDesktopDC)
        
    def __eq__(self, rc):
        if isinstance(rc, Rectangle) and self._rect == rc._rect:
            return True
        else:
            return False
    
    def __ne__(self, rc):
        return not self.__eq__(rc)
        
class ProcessMem(object):
    '''跨进程数据读写
    '''
    def __init__(self, processId, buffer_size=None, remote_buffer=None):
        '''构造函数。如果remote_buffer不为None，直接使用此远程进程的内存块，否则在远程进程中创建一个字节数为buffer_size的内存块。 
        '''
        self._hProcess = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, False, processId).Detach()
        if self._hProcess is None:
            raise RuntimeError("Fail to open process %d" % processId)
        self._release_mem = False
        if remote_buffer is not None:
            self._buffer = remote_buffer
        elif buffer_size is not None:
            self._release_mem = True
            self._buffer = ctypes.windll.kernel32.VirtualAllocEx(self._hProcess, None, buffer_size, win32con.MEM_COMMIT, win32con.PAGE_READWRITE)
            if self._buffer is None:
                raise RuntimeError("Fail to alloc remote memory.")
        else:
            raise ValueError("buffer_size 和 remote_buffer不能同时为空")
    
    def __del__(self):
        if self._release_mem and self._buffer:
            ctypes.windll.kernel32.VirtualFreeEx(self._hProcess, self._buffer, None, win32con.MEM_RELEASE)
            
        if self._hProcess:
            win32api.CloseHandle(self._hProcess)
    
    @property
    def Buffer(self):
        """返回远程进程中内存buffer地址
        """
        return self._buffer
    
    def write(self, local_buffer, buffer_size):
        """将local_buffer中的数据写入远程内存中。
        
        :type local_buffer: buffer
        :param local_buffer: A pointer to the buffer that contains data to be written in the address space of the specified process
        :type buffer_size: unsigned long
        :param buffer_size: The number of bytes to be written to the specified process 
        """
        rst = ctypes.windll.kernel32.WriteProcessMemory(self._hProcess, self._buffer, local_buffer, buffer_size, None)
        if rst == 0:
            raise RuntimeError("Fail to write to remote process memory")

    def read(self, local_buffer, buffer_size):
        """将远程数据读到本地buffer
        
        :type local_buffer: buffer
        :param local_buffer: A pointer to a buffer that receives the contents from the address space of the specified process
        :type buffer_size: unsigned long
        :param buffer_size: The number of bytes to be read from the specified process 
        """
        rst = ctypes.windll.kernel32.ReadProcessMemory(self._hProcess, self._buffer, local_buffer, buffer_size, None)
        if rst == 0:
            raise RuntimeError('Fail to read from remote process memory')
    
#===============================================================================
# 字符串编码相关
#===============================================================================
def getEncoding(s):
    '''获取字符串的编码
    
    :rtype: string
    :return: 'GBK','UNICODE','UTF-8','UNKNOWN'
    '''
    if not isinstance(s, six.string_types):
        return 'UNKNOWN'
    
    if isinstance(s, six.text_type):
        return 'UNICODE'
    
    try:
        unicode(s, 'UTF-8').encode('GBK')
        return 'UTF-8'
    except:
        pass
    
    try:
        unicode(s, 'GBK').encode('UTF-8')
        return 'GBK'
    except:
        pass
        
    try:
        unicode(s, 'UTF-8')
        return 'UTF-8'
    except:
        pass
        
    return 'UNKNOWN'

def myEncode(s, ec='unicode', sc=None):
    '''将字符串转换为指定的编码的字符串
    
    :type s: string
    :param s: 待转换的字符串
    :type ec: string
    :param ec: 待转换的编码. ['GBK','UNICODE','UTF-8']
    :type sc: string
    :param sc: 待转换的字符串的编码
    :attention: sc默认值为None，此时函数会自动判断s的编码（有一定概率会判断错误）
    :return: 转换后的字符串
    '''
    if not isinstance(s, six.string_types):
        return s
    if not sc:
        sc = getEncoding(s)
    else:
        sc = sc.upper()
    ec = ec.upper()
    if sc == 'UTF8':
        sc = 'UTF-8'
    if ec == 'UTF8':
        ec = 'UTF-8'
    validec = ['GBK','UNICODE','UTF-8']
    
    if (ec not in validec) or (sc not in validec) or (sc==ec):
        return s
    
    if ec == 'UNICODE':
        return unicode(s, sc)
    elif ec == 'GBK':
        if sc == 'UTF-8':
            return unicode(s, 'UTF-8').encode('GBK')
        else:
            return s.encode('GBK')
    elif ec == 'UTF-8':
        if sc == 'GBK':
            return unicode(s, 'GBK').encode('UTF-8')
        else:
            return s.encode('UTF-8')
    return s

#===============================================================================
# 获取系统Tips
#===============================================================================
def getToolTips(className='tooltips_class32', retry=10):
    """ 获取系统的浮动tips
    
    :type className: 字符串
    :param className: 类名，默认值为"tooltips_class32"
    :type retry: 整数
    :param retry: 尝试次数，每个0.5秒尝试一次
    """
    from qt4c import qpath
    while retry > 0:
        ctrls = qpath.QPath("/Visible='True'").search()
        for ctrl in ctrls:
            if not ctrl.Valid:
                continue
            if ctrl.ClassName == className:
                tips = ctrl.Caption
                if tips:
                    return tips
        retry = retry - 1
        time.sleep(0.5)
    return ''
            
class MsgSyncer(object):
    '''Thread Post Message 同步
    '''
    pid_event_map = {}
    def __init__(self, hwnd):
        '''Constructor
        
        :param hwnd: 需要同步消息的窗口
        '''
        import qt4c.wincontrols as win
        import qt4c.qpath as qpath
        
        self._syncwnd = 0
        self._eventobj = 0
                
        wndobj = win.Control(root=hwnd)
        try:
            self._pid = wndobj.ProcessId
        except win32gui.error as e:
            if e.winerror == winerror.ERROR_INVALID_WINDOW_HANDLE: #invalid hwnd
                return
            else:
                raise e
        if self._pid == 0: #窗口消失，获取pid为0
            return
        
        if self._pid in MsgSyncer.pid_event_map:
            self._syncwnd, self._eventobj = MsgSyncer.pid_event_map[self._pid]
        else:
            qp = qpath.QPath("/classname='AssistWnd' && processid='%d'" % self._pid)
            wnds = qp.search(None)
            if len(wnds) >1 :
                raise RuntimeError("在一个进程中发现两个SyncWnd")
            elif len(wnds) == 1:
                self._syncwnd = wnds[0].HWnd
                self._eventobj = win32event.CreateEvent(None, True, False, "Event_Sync_%d" % self._pid)
                MsgSyncer.pid_event_map[self._pid] = (self._syncwnd, self._eventobj)
            else:
                import testbase.logger as logger
                logger.debug("Cannot find AssistWnd(%s)" % qp)
    
    def wait(self, timeout=60):
        '''等待消息同步
        '''
        if self._syncwnd == 0:
            return
        win32event.ResetEvent(self._eventobj)
        
        try:
            win32gui.PostMessage(self._syncwnd, win32con.WM_TIMER, 0, 0)
        except pywintypes.error as e:
            if e.winerror == winerror.ERROR_INVALID_WINDOW_HANDLE:
                return
            raise e
        
        start_time = time.time()
        end_time = start_time
        while end_time - start_time <= timeout:
            if win32gui.IsWindow(self._syncwnd):
                ret = win32event.WaitForSingleObject(self._eventobj, 500)
                if ret == win32con.WAIT_OBJECT_0:
                    break
                else:
                    end_time = time.time()
            else:
                win32api.CloseHandle(self._eventobj)
                del MsgSyncer.pid_event_map[self._pid]
                break
        if end_time - start_time > timeout:
            raise RuntimeError('消息同步超时(%d秒)'%timeout)
            
def remote_inject_dll(process_id, dll_path):
    '''在process_id进程中远程注入dll_path的DLL
    '''
    import locale
    os_encoding = locale.getdefaultlocale(None)[1]
    dll_path = dll_path.decode('utf-8').encode(os_encoding)
    
    nbytes = len(dll_path)
    dllpath = str(dll_path)
    hproc = ctypes.windll.kernel32.OpenProcess(win32con.PROCESS_ALL_ACCESS, False, process_id)
    dllname = ctypes.windll.kernel32.VirtualAllocEx(hproc, None, nbytes, 
                            win32con.MEM_RESERVE|win32con.MEM_COMMIT, win32con.PAGE_READWRITE)
    ctypes.windll.kernel32.WriteProcessMemory(hproc, dllname, dllpath, len(dllpath), None)
    hknl = ctypes.windll.kernel32.GetModuleHandleA("kernel32.dll")
    ldlib_addr = ctypes.windll.kernel32.GetProcAddress(hknl , "LoadLibraryA")
    ctypes.windll.kernel32.CreateRemoteThread(hproc, None, None, 
                ldlib_addr, dllname, None, None)
    ctypes.windll.kernel32.VirtualFreeEx(hproc, dllname, nbytes, win32con.MEM_RELEASE)
    ctypes.windll.kernel32.CloseHandle(hknl)
    ctypes.windll.kernel32.CloseHandle(hproc)

class Process(object):
    '''
    提供系统进程的操作类
    使用例子：
    for proc in Process.GetProcessesByName('iexplore.exe'):
        print proc.ProcessId
    '''
    
    def __init__(self, pid):
        '''构造函数
        
        :param pid: 进程ID 
        '''
        self._pid = pid
        
    @staticmethod
    def GetProcessesByName(process_name):
        '''返回进程名为process_name的Process类实例列表
        '''
        if six.PY3:
            process_name = process_name.encode('utf8')
        processes = []
        
        TH32CS_SNAPPROCESS = 0x00000002
        pe = wintypes.PROCESSENTRY32() 
        hSnapshot = ctypes.windll.kernel32.CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0)
        pe.dwSize = ctypes.sizeof(wintypes.PROCESSENTRY32)
        if not ctypes.windll.kernel32.Process32First(hSnapshot, ctypes.byref(pe)):
            win32api.CloseHandle(hSnapshot)
            return processes
        
        while 1:
            if pe.szExeFile.upper() == process_name.upper():
                processes.append(Process(pe.th32ProcessID))
            
            if not ctypes.windll.kernel32.Process32Next(hSnapshot, ctypes.byref(pe)):
                break
            
        win32api.CloseHandle(hSnapshot)
        return processes
    
    @property
    def ProcessName(self):
        # 64位python下有问题
        '''返回进程名字。失败返回None
        '''
        TH32CS_SNAPPROCESS = 0x00000002
        pe = wintypes.PROCESSENTRY32() 
        hSnapshot = ctypes.windll.kernel32.CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0)
        pe.dwSize = ctypes.sizeof(wintypes.PROCESSENTRY32)
        if not ctypes.windll.kernel32.Process32First(hSnapshot, ctypes.byref(pe)):
            win32api.CloseHandle(hSnapshot)
            return None
        
        while 1:
            if pe.th32ProcessID == self._pid:
                win32api.CloseHandle(hSnapshot)
                if six.PY3:
                    return pe.szExeFile.decode('utf-8')
                return pe.szExeFile
            
            if not ctypes.windll.kernel32.Process32Next(hSnapshot, ctypes.byref(pe)):
                break
            
        win32api.CloseHandle(hSnapshot)
        return None
    
    @property
    def Live(self):
        return (win32process.GetProcessVersion(self._pid) > 0)
    
    @property
    def ProcessId(self):
        return self._pid
    
    def waitForQuit(self, timeout=10, interval=0.5):
        """在指定的时间内等待退出
        
        :return: 如果在指定的时间内退出，返回True；否则返回False
        """
        if not self.Live:
            return True
        
        try: 
            Timeout(timeout, interval).waitObjectProperty(self, "Live", False)
        except TimeoutError:
            return False
        return True
    
    def _adjustProcessPrivileges(self):
        """提升权限
        """
        import win32security
        hCurPro = win32process.GetCurrentProcess()
        hToken = win32security.OpenProcessToken(hCurPro, win32security.TOKEN_ADJUST_PRIVILEGES | win32security.TOKEN_QUERY)
        luid = win32security.LookupPrivilegeValue(None, win32security.SE_DEBUG_NAME)
        if luid:
            newState = [(luid, win32security.SE_PRIVILEGE_ENABLED),]
            win32security.AdjustTokenPrivileges(hToken, 0, newState)
        win32api.CloseHandle(hToken)
    
    def terminate(self):
        """终止进程
        """
        from qt4c.wincontrols import TrayNotifyBar

        self._adjustProcessPrivileges()
        try:
            hProcess = win32api.OpenProcess(win32con.PROCESS_TERMINATE, False, self._pid).Detach()
            win32process.TerminateProcess(hProcess, 0)
            win32api.CloseHandle(hProcess)
            tb = TrayNotifyBar()
            item = tb[self._pid]
            if item:
                item.destroy()
        except:
            pass
    
    @property
    def ProcessPath(self):
        """获取进程可执行文件的全路径
        """
        self._adjustProcessPrivileges()
        try:
            hproc=win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS,False,self._pid)
            exe_path=win32process.GetModuleFileNameEx(hproc,0)
            return exe_path
        except WindowsError:
            return "(failed to get)"
        
def is_system_locked():
    '''检测系统是否处于锁屏界面
    
    @return: bool 如果处于锁屏，返回True
    '''

    try:
        UserDll=ctypes.WinDLL("User32.dll")
        hDesk=UserDll.OpenDesktopW(u"Default", 0, False, DESKTOP_SWITCHDESKTOP)
#         PyhDesk=win32service.OpenDesktop("Default",0,FALSE,DESKTOP_SWITCHDESKTOP)
        bUnlocked = UserDll.SwitchDesktop(hDesk)
        UserDll.CloseDesktop(hDesk)
    except win32service.error as e:
        print(str(e))
        bUnlocked=False
    
    return (not bUnlocked)

class MetisView(object):
    '''各端实现的MetisView
    '''
    def __init__(self, control):
        self._control = control

    @property
    def rect(self):
        '''元素相对坐标(x, y, w, h)
        '''
        x = 0
        y = 0
        w = abs(self._control.BoundingRect.Right - self._control.BoundingRect.Left)
        h = abs(self._control.BoundingRect.Top - self._control.BoundingRect.Bottom)
        return x, y, w, h

    @property
    def os_type(self):
        '''系统类型，例如"android"，"ios"，"pc"
        '''
        return 'pc'

    def screenshot(self):
        '''当前容器的区域截图
        '''
        
        bbox =(self._control.BoundingRect.Left,self._control.BoundingRect.Top,
               self._control.BoundingRect.Right,self._control.BoundingRect.Bottom)
        im = ImageGrab.grab(bbox)
        return im

    def click(self, offset_x=None, offset_y=None):
        '''点击
        
        :param offset_x: 相对于该控件的坐标offset_x，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_x: float|None
        :param offset_y: 相对于该控件的坐标offset_y，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_y: float|None
        '''
        if offset_x != None:
            offset_x = int(offset_x*abs(self._control.BoundingRect.Right - self._control.BoundingRect.Left))
        if offset_y != None:
            offset_y = int(offset_y*abs(self._control.BoundingRect.Top - self._control.BoundingRect.Bottom))
        self._control.click(xOffset=offset_x, yOffset=offset_y)

    def send_keys(self, text):
        Keyboard.inputKeys(text)
        

    def double_click(self, offset_x=None, offset_y=None):
        pass

    def long_click(self, offset_x=None, offset_y=None):
        pass

def getDpi(hwnd=None):
    if not hwnd:
        hwnd = win32gui.GetDesktopWindow()
    ratio = 1.0
    try:
        # win10 1607版本以上
        dpi = ctypes.windll.user32.GetDpiForWindow(hwnd)
        ratio = dpi / 96.0
    except:
        pass
    return ratio

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
import ctypes
import win32con
import os

if ctypes.sizeof(ctypes.c_void_p) == 8:
    ULONG_PTR = ctypes.c_ulonglong
else:
    ULONG_PTR = ctypes.c_ulong
    
class RECT(ctypes.Structure):
    """The RECT structure defines the coordinates of the upper-left and lower-right corners of a rectangle
    """
    _fields_ = [
        ('left', ctypes.c_long),
        ('top', ctypes.c_long),
        ('right', ctypes.c_long),
        ('bottom', ctypes.c_long)]
    
class TBBUTTON(ctypes.Structure):
    """Contains information about a button in a system's traynotifybar."""
    _fields_ = [
        ('iBitmap', ctypes.c_int),
        ('idCommand', ctypes.c_int),
        ('fsState', ctypes.c_ubyte),
        ('fsStyle', ctypes.c_ubyte),
        ('bReserved', ctypes.c_ubyte*(6 if 'PROGRAMFILES(X86)' in os.environ else 2)),
        ('dwData', ctypes.c_ulong),
        ('iString', ctypes.c_int)]
       
class TRAYDATA(ctypes.Structure):
    """for description"""
    _fields_ = [
        ('hwnd', ctypes.c_long),
        ('uID', ctypes.c_uint),
        ('uCallbackMessage', ctypes.c_uint),
        ('Reserved', ctypes.c_ulong * 2),
        ('hIcon', ctypes.c_long)]
    
class PROCESSENTRY32(ctypes.Structure):
    """
    
    :desc: Describes an entry from a list that enumerates the processes residing in the system address space when a snapshot was taken.
    """
    _fields_ = [
        ('dwSize', ctypes.c_ulong),
        ('cntUsage', ctypes.c_ulong),
        ('th32ProcessID', ctypes.c_ulong),
        ('th32DefaultHeapID', ULONG_PTR),
        ('th32ModuleID', ctypes.c_ulong),
        ('cntThreads', ctypes.c_ulong),
        ('th32ParentProcessID', ctypes.c_ulong),
        ('pcPriClassBase', ctypes.c_long),
        ('dwFlags', ctypes.c_ulong),
        ('szExeFile', ctypes.c_char * win32con.MAX_PATH),
        ('th32MemoryBase', ctypes.c_ulong),
        ('th32AccessKey', ctypes.c_ulong)]
    
class APPBARDATA(ctypes.Structure):
    """App Bar Data Structure
    """
    _fields_ = [
        ('cbSize', ctypes.c_ulong),
        ('hWnd', ctypes.c_ulong),
        ('uCallbackMessage', ctypes.c_ulong),
        ('uEdge', ctypes.c_ulong),
        ('rc', RECT),
        ('lParam', ctypes.c_long)]
    
class LVITEM(ctypes.Structure):
    """
    
    :desc: Specifies or receives the attributes of a list-view item. This structure has been updated
           to support a new mask value (LVIF_INDENT) that enables item indenting. This structure supersedes the LV_ITEM structure
    """
    _fields_ = [
        ('mask', ctypes.c_uint),
        ('iItem', ctypes.c_int),
        ('iSubItem', ctypes.c_int),
        ('state', ctypes.c_uint),
        ('stateMask', ctypes.c_uint),
        ('pszText', ctypes.c_char_p),
        ('cchTextMax', ctypes.c_int),
        ('iImage', ctypes.c_int),
        ('lParam', ctypes.c_char_p),
        ('iIndent', ctypes.c_int),
        ('iGroupId', ctypes.c_int),
        ('cColumns', ctypes.c_int),
        ('puColumns', ctypes.c_uint)]
    
class LVITEM64(ctypes.Structure):
    """
    
    :desc: Specifies or receives the attributes of a list-view item. This structure has been updated
           to support a new mask value (LVIF_INDENT) that enables item indenting. This structure supersedes the LV_ITEM structure
    """
    _fields_ = [
        ('mask', ctypes.c_uint),
        ('iItem', ctypes.c_int),
        ('iSubItem', ctypes.c_int),
        ('state', ctypes.c_uint),
        ('stateMask', ctypes.c_uint),
        ('iPlaceholder11', ctypes.c_int),
        ('pszText', ctypes.c_char_p),
        ('iPlaceholder12', ctypes.c_int),
        ('cchTextMax', ctypes.c_int),
        ('iImage', ctypes.c_int),
        ('lParam', ctypes.c_char_p),
        ('iPlaceholder2', ctypes.c_int),
        ('iIndent', ctypes.c_int),
        ('iGroupId', ctypes.c_int),
        ('cColumns', ctypes.c_int),
        ('iPlaceholder3', ctypes.c_int),
        ('puColumns', ctypes.c_uint),
        ('iPlaceholder4', ctypes.c_int),
        ]
    
class TVITEM(ctypes.Structure):
    """
    """
    _fields_ = [
        ('mask', ctypes.c_uint),
        ('hItem', ctypes.c_long),
        ('state', ctypes.c_uint),
        ('stateMask', ctypes.c_uint),
        ('pszText', ctypes.c_char_p),
        ('cchTextMax', ctypes.c_int),
        ('iImage', ctypes.c_int),
        ('iSelectedImage', ctypes.c_int),
        ('cChildren', ctypes.c_int),
        ('lParam', ctypes.c_char_p)]
    
class BITMAPINFOHEADER(ctypes.Structure):
    """BITMAP Info Header
    """
    _fields_ = [
        ('biSize', ctypes.c_uint32),
        ('biWidth', ctypes.c_long),
        ('biHeight', ctypes.c_long),
        ('biPlanes', ctypes.c_uint16),
        ('biBitCount', ctypes.c_uint16),
        ('biCompression', ctypes.c_uint32),
        ('biSizeImage', ctypes.c_uint32),
        ('biXPelsPerMeter', ctypes.c_long),
        ('biYPelsPerMeter', ctypes.c_long),
        ('biClrUsed', ctypes.c_uint32),
        ('biClrImportant', ctypes.c_uint32)
                ]

class RGBTRIPLE(ctypes.Structure):
    """RGB Define
    """
    _fields_ = [
        ('rgbBlue', ctypes.c_byte),
        ('rgbGreen', ctypes.c_byte),
        ('rgbRed', ctypes.c_byte),
        ('rgbReserved', ctypes.c_byte),
                ]
        
class BITMAPINFO(ctypes.Structure):
    """BITMAP Info
    """
    _fields_ = [
        ('bmiHeader', BITMAPINFOHEADER),
        ('bmciColors', RGBTRIPLE * 1)]
  
class BITMAP(ctypes.Structure):
    """BITMAP
    """
    _fields_ = [
        ('bmType', ctypes.c_long),
        ('bmWidth', ctypes.c_long),
        ('bmHeight', ctypes.c_long),
        ('bmWidthBytes', ctypes.c_long),
        ('bmPlanes', ctypes.c_uint16),
        ('bmBitsPixel', ctypes.c_uint16),
        ('bmBits', ctypes.c_void_p)]
      
class DIBSECTION(ctypes.Structure):
    """DIB Section
    """
    _fields_ = [
        ('dsBm', BITMAP),
        ('dsBmih', BITMAPINFOHEADER),
        ('dsBitfields', ctypes.c_uint32 * 3),
        ('dshSection', ctypes.c_void_p),
        ('dsOffset', ctypes.c_uint32)]
    
class BITMAPFILEHEADER(ctypes.Structure):
    """BITMAP File Header
    """
    _fields_ = [
        ('bfType', ctypes.c_uint16),
        ('bfSize', ctypes.c_uint32),
        ('bfReserved1', ctypes.c_uint16),
        ('bfReserved2', ctypes.c_uint16),
        ('bfOffBits', ctypes.c_uint32)
                ]
    
class MODULEENTRY32(ctypes.Structure):
    """This structure describes an entry from a list that enumerates 
    the modules used by a specified process.
    """
    _fields_ = [
        ("dwSize",ctypes.c_long),
        ("th32ModuleID", ctypes.c_long),
        ("th32ProcessID", ctypes.c_long),
        ("GlblcntUsage", ctypes.c_long),
        ("ProccntUsage", ctypes.c_long),
        ("modBaseAddr", ctypes.c_long),
        ("modBaseSize", ctypes.c_long),
        ("szModule", ctypes.c_char*win32con.MAX_PATH),
        ("szExePath", ctypes.c_char*win32con.MAX_PATH),
        ("dwFlags", ctypes.c_long),
         ]

class THREADENTRY32(ctypes.Structure):
    """
    This structure describes an entry from a list that enumerates the threads executing in the system when a snapshot was taken. 
    """
    _fields_ = [
        ("dwSize", ctypes.c_ulong),
        ("cntUsage", ctypes.c_ulong),
        ("th32ThreadID", ctypes.c_ulong),
        ("th32OwnerProcessID", ctypes.c_ulong),
        ("tpBasePri", ctypes.c_long),
        ("tpDeltaPri", ctypes.c_long),
        ("dwFlags", ctypes.c_ulong)
    ]



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''IE WebView实现
'''

import time
import logging
import win32gui
import win32process

from qt4c.webview.iewebview.iedriver import IEDriver
from qt4w.util import JavaScriptError
from qt4c.webview.base import WebViewBase
from qt4c.mouse import Mouse, MouseFlag, MouseClickType

class IEWebView(WebViewBase):
    '''IE WebView实现
    '''
    def __init__(self, ie_window_or_hwnd):
        '''初始化
        
        :params ie_window_or_hwnd: ie窗口或句柄
        :type ie_window_or_hwnd: Control or int
        '''
        if isinstance(ie_window_or_hwnd, int):  # 句柄需要转化为对应的窗口，句柄是IEFrame的句柄
            process_id = win32process.GetWindowThreadProcessId(ie_window_or_hwnd)[1]
            from browser.ie import IEWindow
            ie_window = IEWindow(process_id).ie_window
        else:
            ie_window = ie_window_or_hwnd
        from qt4w.webdriver import iewebdriver
        self._webdriver = iewebdriver.IEWebDriver(self)
        self._browser_type = 'ie'
        self._ie_window = ie_window
        self._driver = IEDriver(self._ie_window.HWnd)
        super(IEWebView, self).__init__(ie_window, self._webdriver)
        
    def _get_frame_window_by_xpath(self, frame_xpaths):
        '''根据xpath查找对应的frameIHTMLWindow对象
        '''
        if len(frame_xpaths) == 0: return None
        else:
            parent_frame_win = self._get_frame_window_by_xpath(frame_xpaths[:-1])
            frame_id, url = self._webdriver._get_frame_info(frame_xpaths)
            # print frame_id, url
            return self._driver.get_frame_window(parent_frame_win, frame_id, url)
    
    def eval_script(self, frame_xpaths, script, use_eval=True):
        '''在指定frame中执行JavaScript，并返回执行结果
        
        :param frame_xpaths: frame元素的XPATH路径，如果是顶层页面，怎传入“[]”
        :type frame_xpaths:  list
        :param script:       要执行的JavaScript语句
        :type script:        string
        '''
        import pywintypes
        from qt4c.webview.iewebview.iedriver import IEDriverError
        frame_win = None
        retry_count = 3
        for i in range(retry_count):
            try:
                # print 'xxx', frame_xpaths
                if frame_xpaths:
                    frame_win = self._get_frame_window_by_xpath(frame_xpaths)
                result = self._driver.eval_script(frame_win, script, use_eval)
                break
            except (pywintypes.com_error, IEDriverError) as e:
                if i >= retry_count - 1:
                    raise e
                logging.exception('eval script error')
                time.sleep(1)
                self._driver._init_com_obj()  # 重新初始化COM对象
        if result == None: return
        return self._handle_result(result, frame_xpaths)
    
    def highlight(self, elem_xpaths):
        '''使元素高亮
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        '''
        try:
            return self._webdriver.highlight(elem_xpaths)
        except JavaScriptError as e:
            # IE下可能会出现setAttribute时拒绝访问的错误，此时需要重新创建DIV
            logging.warn('highlight error: %s' % e)
            self.eval_script(elem_xpaths[:-1], 'qt4w_driver_lib.initHighlightDiv();')
            return self._webdriver.highlight(elem_xpaths)
        
    # 因为ie在使用Mouse.postClick情况下，需要发送先发送两次WM_LBUTTONDOWN，然后发送WM_LBUTTONUP才能点击生效，
    # 为了不修改qt4c的内容，所以这里使用全局的鼠标点击实现，但是这种情况下，如果页面被遮挡，点击仍然会失败
    # def inner_click(self, flag, click_type, x_offset, y_offset):
    #     new_x, new_y = win32gui.ClientToScreen(self._window.HWnd, (int(x_offset), int(y_offset)))
    #     Mouse.click(new_x, new_y, flag, click_type)
    
    # def click(self, x_offset, y_offset):
    #     self.inner_click(MouseFlag.LeftButton, MouseClickType.SingleClick, x_offset, y_offset)
        
    # def double_click(self, x_offset, y_offset):
    #     self.inner_click(MouseFlag.LeftButton, MouseClickType.DoubleClick, x_offset, y_offset)
    
    # def right_click(self, x_offset, y_offset):
    #     self.inner_click(MouseFlag.RightButton, MouseClickType.SingleClick, x_offset, y_offset)
        
    
if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''应用程序基类模块
'''

#PROCESS_TERMINATE = 1


class App(object):
    '''应用程序基类
    '''
    _totalapps = [] #所有打开的应用程序
    def __init__(self):
        '''构造函数，将self加入到应用程序列表中
        '''
        App._totalapps.append(self)
        
    def quit(self):
        '''请在子类中实现，并调用此方法通知程序退出
        '''
        if self in App._totalapps:
            App._totalapps.remove(self)
    
    @staticmethod
    def quitAll():
        '''退出所有应用程序
        '''
        apps = App._totalapps[:]
        for app in apps:
            app.quit()
            
    @staticmethod
    def clearAll():
        '''清除所有程序记录
        '''
        del App._totalapps[:]
        
    @staticmethod
    def killAll():
        '''结束所有记录的进程
        '''
        from util import Process
        for app in App._totalapps:
            Process(app.ProcessId).terminate()
        del App._totalapps[:]
        
        
    

if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#

'''ChromeBrowser的接口实现
'''

import logging, os, re, time
import shutil
import win32api
import win32com.client
import win32event
import win32gui
import six

from qt4w.browser.browser import IBrowser
from qt4c.qpath import QPath
from qt4c.webview.chromewebview import ChromeWebView
from qt4c.webview.chromewebview.chromedriver import ChromeDriver
from qt4c.wincontrols import Window


class ChromeBrowser(IBrowser):
    '''Chrome浏览器
    '''
    temp_path = os.environ['temp']
    
    def __init__(self, port=9200):
        self._port = port
        self._port_list = [port]
        self._main_wnd = None
        self._pid = 0
        self._temp_dir = os.path.join(self.__class__.temp_path, 'Chrome_%d')
    
    def _handle_title(self, title):
        '''处理标题
        '''
        return title.replace('&amp;', '&')
        
    def get_chrome_window_list(self, pid):
        '''通过pid查找对应的chrome窗口列表
        '''
        from qt4c.exceptions import ControlAmbiguousError
        qp_str = r"/ClassName='Chrome_WidgetWin_1' && ProcessId='%d' && Visible='True' %%s /ClassName='Chrome_RenderWidgetHostHWND' && Visible='True'" % pid
        try:
            win = Window(locator=QPath(qp_str % ''))
            if win.exist() and win.Valid:  # 这句话才会真正去查找控件;增加有效性判断
                return [win]
            else:
                return []
        except ControlAmbiguousError as e:
            pattern = re.compile(r'找到(\d+)个控件')
            ret = pattern.search(str(e))
            if not ret: raise
            win_num = int(ret.group(1))
            win_list = []
            for i in range(win_num):
                win = Window(locator=QPath(qp_str % ("&& Instance='%d'" % i)))
                if win.exist(): win_list.append(win)
            return win_list
        
    def open_url(self, url, page_cls=None):
        import win32process, win32con
        while is_port_occupied(self._port):  # 如果端口被占用，则查找下一个可用端口
            self._port += 1
        if not self._port in self._port_list:
            self._port_list.append(self._port)
        temp_dir = self._temp_dir % self._port
        exe_path = ChromeBrowser.get_browser_path()
        params = "--remote-debugging-port=%d --disable-session-crashed-bubble --disable-features=TranslateUI --disable-breakpad --no-first-run --new-window --disable-desktop-notifications --user-data-dir=%s" % (self._port, temp_dir)
        cmd = ' '.join([exe_path, params, url])
        logging.debug('chrome: %s' % cmd)
        _, _, pid, _ = win32process.CreateProcess(None, cmd, None, None, 0, 0, None, None, win32process.STARTUPINFO())
        handle = win32api.OpenProcess(win32con.PROCESS_QUERY_INFORMATION, False, pid)
        win32event.WaitForInputIdle(handle, 10000)
        self._pid = pid
        logging.info('chrome进程为%d' % pid)  # 加上此句话查看chrome是否成功打开了
        timeout = 10
        time0 = time.time()
        while time.time() - time0 < timeout:
            win_list = self.get_chrome_window_list(pid)
            if len(win_list) > 0: break
            time.sleep(1)
        else:
            raise RuntimeError('find chrome browser window failed')
        assert(len(win_list) == 1)
        page_wnd = win_list[0]
        self._webview = ChromeWebView(page_wnd, url, self._pid, self._port)
        return self.get_page_cls(self._webview, page_cls)
   
    def find_by_url(self, url, page_cls=None, timeout=10):
        '''在当前打开的页面中查找指定url,返回WebPage实例，如果未找到，则抛出异常
        '''
        from qt4c.exceptions import TimeoutError
        time0 = time.time()
        while time.time() - time0 < timeout:
            try:
                webview = self.search_chrome_webview(url)
                break
            except RuntimeError as e:
                logging.warn('[ChromeBrowser] search chrome window failed: %s' % e)
                time.sleep(0.5)
        else:
            raise TimeoutError
        return self.get_page_cls(webview, page_cls)
     
    def get_page_cls(self, webview, page_cls=None):
        '''得到具体页面类
        '''
        if page_cls:
            return page_cls(webview)
        return webview
    
    @staticmethod
    def get_browser_path():
        '''获取chorme.exe的路径
        '''
        from six.moves import winreg
        sub_key = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\Google Chrome"
        install_dir = ''
        try:
            hkey = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, sub_key)
            install_dir, _ = winreg.QueryValueEx(hkey, "InstallLocation")
            winreg.CloseKey(hkey)
        except WindowsError:
            try:
                hkey = winreg.OpenKey(winreg.HKEY_CURRENT_USER, sub_key)
                install_dir, _ = winreg.QueryValueEx(hkey, "InstallLocation")
            except WindowsError:
                try:
                    sub_key = r"SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\chrome.exe"
                    hkey = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, sub_key)
                    install_dir, _ = winreg.QueryValueEx(hkey, "Path")
                except WindowsError:
                    for env_dir in os.getenv('PATH').split(';'):
                        if os.path.exists(os.path.join(env_dir, 'chrome.exe')):
                            install_dir = env_dir
                            break
                    if not install_dir:
                        app_dirs = [os.getenv('LOCALAPPDATA'), os.getenv('ProgramFiles(x86)'), os.getenv('ProgramW6432')]
                        for app_dir in app_dirs:
                            if(os.path.exists(os.path.join(app_dir, r'Google\Chrome\Application', 'chrome.exe'))):
                                install_dir = os.path.join(app_dir, r'Google\Chrome\Application')
                                break
        if not install_dir:
            raise WindowsError('未找到Chrome可执行文件路径')
        return os.path.join(install_dir, "chrome.exe")
    
    def search_chrome_webview(self, url):
        '''根据url查找chrome对应的webview类
        
        returns ChromeWebView: ChromeWebView类
        '''
        chrome_path = self.get_browser_path()
        for port in self._port_list:
            chrome_driver = ChromeDriver(port)
            title = ''
            for it in chrome_driver.get_page_list():
                if it['url'] == url or re.match(url, it['url']):
                    title = self._handle_title(it['title'])
                    url = it['url']
                    if six.PY2 and isinstance(title, six.text_type): title = title.encode('utf8')
                    break
            else: continue
            break
        else:
            raise RuntimeError('获取页面：%s 标题失败' % url)
        
        wmi = win32com.client.GetObject('winmgmts:')
        for p in wmi.InstancesOf('win32_process'):
            if not p.CommandLine:
                continue
            
            if p.CommandLine.startswith(chrome_path) or p.CommandLine.startswith('"%s"' % chrome_path):
                items = p.CommandLine[len(chrome_path):].split()
                if not items or items[0].startswith('--type='): continue  # 都不是Browser进程
                
                chrome_window_list = self.get_chrome_window_list(p.ProcessId)
                for chrome_window in chrome_window_list:
                    win_title = chrome_window.Parent.Text
                    if win_title.endswith(' - Google Chrome'): win_title = win_title[:-16]

                    if win_title == title:
                        try:
                            win32gui.SetForegroundWindow(chrome_window.TopLevelWindow.HWnd)
                        except Exception as e:
                            print(e)
                            logging.warn('win32gui.SetForegroundWindow error')
                        self._webview = ChromeWebView(chrome_window, url, p.ProcessId)
                        return self._webview
                raise RuntimeError('当前标签页不在窗口最前端！')
        else:
            raise RuntimeError('%s对应的chrome进程不存在' % url)


    def clear_cache(self):
        for p in self._port_list:
            temp_dir = self._temp_dir % p
            if os.path.isdir(temp_dir):
                logging.info('clear dir %s' % temp_dir)
                try:
                    shutil.rmtree(temp_dir)
                except:
                    logging.error('clear dir %s failed' % temp_dir)

    def close(self):
        chrome_path = self.get_browser_path()
        wmi = win32com.client.GetObject('winmgmts:')
        for p in wmi.InstancesOf('win32_process'):
            if not p.CommandLine:
                continue
            
            if p.CommandLine.startswith(chrome_path) or p.CommandLine.startswith('"%s"' % chrome_path):
                items = p.CommandLine[len(chrome_path):].split()
                if not items or items[0].startswith('--type='): continue  # 都不是Browser进程
                for item in items:
                    if item.find('=') != -1:
                        [commandType, commandValue] = item.split('=', 1)
                        if(commandType == '--remote-debugging-port' and int(commandValue) in self._port_list):
                            try:
                                p.Terminate
                                time.sleep(0.5) # 等待进程完全退出
                                self.clear_cache()
                                break
                            except Exception as e:
                                logging.error('kill process failed, error: %s' % e)
    
    @staticmethod
    def killall():
        '''杀掉所有chrome进程
        '''
        wmi = win32com.client.GetObject('winmgmts:')
        for p in wmi.InstancesOf('win32_process'):
            if p.Name == 'chrome.exe':
                try:
                    p.Terminate
                except:
                    pass

        for it in os.listdir(ChromeBrowser.temp_path):
            p = os.path.join(ChromeBrowser.temp_path, it)
            if os.path.isdir(p) and it.startswith('Chrome_'):
                logging.info('clear dir %s' % p)
                try:
                    shutil.rmtree(p)
                except:
                    logging.error('clear dir %s failed' % p)
        
    @property
    def Url(self):
        return self._webview.eval_script(None, 'location.href')
 
def is_port_occupied(port):
    '''
    端口是否被占用
    '''
    import socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.bind(('localhost', port))
        sock.close()
        return False
    except:
        return True
    
def get_next_avail_port(port):
    cur_port = port
    while is_port_occupied(cur_port):
        cur_port += 1
    return cur_port
    
if __name__ == '__main__':
    pass
    


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''QT4C(Client Driver for QTA)
'''



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#

'''Chrome WebView实现
'''

import time
import logging
from qt4c.webview.base import WebViewBase
from qt4c import qpath
from qt4w.webdriver.webkitwebdriver import WebkitWebDriver
from qt4c.webview.chromewebview.chromedriver import ChromeDriver


class ChromeWebView(WebViewBase):
    '''Chrome WebView实现
    '''
    
    def __init__(self, page_wnd, url, pid, port=9200):
        self._port = port
        self._url = url
        self._pid = pid
        self._window = page_wnd
        super(ChromeWebView, self).__init__(self._window, WebkitWebDriver(self))
        self._driver = ChromeDriver(self._port).get_debugger(self._url)
        self._frame_dict = {}  # 缓存frame xpath和frame id的对应关系
        self._browser_type = 'chrome'
    
    def _get_frame(self, parent, name, url):
        '''根据frame的name和url获取frameTree节点
        
        :param parent 要查找的frameTree节点
        :type parent  dict
        :param name: frame的id或name属性
        :type name:  string
        :param url:  frame的url
        :type url:   string
        '''
        if 'childFrames' in parent:
            for child in parent['childFrames']:
                if (name and child['frame']['name'] == name) or (url and child['frame']['url'] == url):
                    return child
        else:
            logging.info('[ChromeWebView] get_frame %s' % parent)
        # raise RuntimeError('find frame of %s failed' % ((('name=%s' % name) if name else ('url=%s' % url))))
        return None
    
    def _get_frame_id_by_xpath(self, frame_xpaths):
        '''根据XPath对象查找frame id
        
        :param frame_xpaths: frame的xpath数组
        :type frame_xpaths: list
        '''
        frame_xpaths_str = ''.join(frame_xpaths)
        if frame_xpaths_str in self._frame_dict: return self._frame_dict[frame_xpaths_str]
        # 缓存中不存在时
        timeout = 10
        time0 = time.time()
        while time.time() - time0 < timeout:
            frame_tree = self._driver.get_frame_tree()
            frame_id = frame_tree['frame']['id']
            if len(frame_xpaths) == 0: 
                # 顶层frame
                self._frame_dict[frame_xpaths_str] = frame_id
            else:
                frame_exist = True
                for i in range(len(frame_xpaths)):
                    name, url = self._webdriver._get_frame_info(frame_xpaths[:i + 1])
                    frame_tree = self._get_frame(frame_tree, name, url)
                    if frame_tree == None:
                        # 未找到frame
                        frame_exist = False
                        break
                    self._frame_dict[frame_xpaths_str] = frame_tree['frame']['id']
                if not frame_exist: 
                    time.sleep(0.5)
                    continue
            return self._frame_dict[frame_xpaths_str]
    
    def eval_script(self, frame_xpaths, script):
        '''在指定frame中执行JavaScript，并返回执行结果
        
        :param frame_xpaths: frame元素的XPATH路径，如果是顶层页面，则传入“[]”或者是frame id
        :type frame_xpaths:  list or string
        :param script:       要执行的JavaScript语句
        :type script:        string
        '''
        from qt4c.webview.chromewebview.chromedriver import ChromeDriverError
        timeout = 10
        time0 = time.time()
        while time.time() - time0 < timeout:
            if isinstance(frame_xpaths, list):
                frame_id = self._get_frame_id_by_xpath(frame_xpaths)
            else:
                frame_id = frame_xpaths
            try:
                result = self._driver.eval_script(frame_id, script)
                break
            except (ChromeDriverError, RuntimeError) as e:
                if isinstance(e, RuntimeError):
                    del self._frame_dict[''.join(frame_xpaths)]
                    time.sleep(0.5)
                elif e.code == -32000:
                    # Execution context with given id not found.
                    time.sleep(0.5)
                else:
                    raise e
        else:
            raise RuntimeError('执行JavaScript代码失败')
        return self._handle_result(result, frame_xpaths)
    
    def click(self, x_offset, y_offset): 
        '''Chrome中按住shift键点击，以便在新窗口中打开页面
        '''
        from qt4c.keyboard import Key
        js_tmpl = '''(function(x, y){
    var getNodeByPos = function(tag, x, y){
        var nodes = document.getElementsByTagName(tag);
        for(var i=0;i<nodes.length;i++){
            var rect = nodes[i].getBoundingClientRect();
            if(rect.left <= x && rect.right >= x && rect.top <= y && rect.bottom >= y) return nodes[i];
        }
        return null;
    };
    
    var getTargetFrame = function(x, y){
        var frame = getNodeByPos('iframe', x, y);
        if(frame != null){
            // target node is in iframe
            var result = new Array();
            var rect = frame.getBoundingClientRect();
            var scale = qt4w_driver_lib.getScale();
            result.push(rect.left*scale);
            result.push(rect.top*scale);
            var name = frame.getAttribute('name');
            if(name != null){
                result.push('//iframe[@name="'+name+'"]');
                return result.toString();
            }
            var id = frame.getAttribute('id');
            if(id != null){
                result.push('//iframe[@id="'+id+'"]');
                return result.toString();
            }
            var src = frame.getAttribute('src');
            if(src == '' || src == 'about:blank') return '';
            result.push('//iframe[@src="' + src + '"]');
            return result.toString();
        }
        return '';
    };

    
    var isButton = function(x,y){
        var button = getNodeByPos('button',x,y)
        return (button != null)  
    };
    
    var isNewTabLinkNode = function(x, y){
        var frame_info = getTargetFrame(x, y);
        if(frame_info != '') return frame_info;
        
        var node = getNodeByPos('a', x, y);
        if(node == null) return null;
        //console.log(node.outerHTML);
        return node.getAttribute('target') == '_blank';
    };
    var scale = qt4w_driver_lib.getScale();
    return isNewTabLinkNode(x/scale, y/scale)||isButton(x/scale, y/scale);
})(%f, %f);'''
        x = x_offset
        y = y_offset
        frame_xpaths = []
        new_tab_flag = False
        while True:
            result = self.eval_script(frame_xpaths, js_tmpl % (x, y))
            if ',' in result:
                # in frame
                result = result.split(',')
                frame_xpaths.append(result[2])
                x -= float(result[0])
                y -= float(result[1])
            else:
                new_tab_flag = result == 'true'
                break
        shift_key = Key(16)  # SHIFT
        if new_tab_flag: shift_key._inputKey(False)
        super(ChromeWebView, self).click(x_offset, y_offset)
        if new_tab_flag: shift_key._inputKey(True)
        
def _get_pid_by_port(port):
    '''利用端口，获取对应端口的进程id
    '''
    
    import subprocess
    import time
    import re
    timeout = 10
    start = time.time()
    cur = start
    command = "netstat -aon | findstr %d" % port
    sub_proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    while 1:
        ret = sub_proc.poll()
        if ret is None:
            time.sleep(0.2)
            cur = time.time()
            if(cur > start + timeout):
                sub_proc.kill()
                break
        elif ret == 0:
            break
        else:
            break
    content = sub_proc.stdout.read().rstrip()
#     print content
    result = re.search('LISTENING\s+(\d+)$', content)
    if result:
        content = result.group(0)
        result = re.search('\d+$', content)
        return int(result.group(0))
    else:
        raise RuntimeError("The specified port %d does not bind to a process" % port)
    
def get_pid_by_port(port):
    '''增加延时和重试机制，防止网络初始化太慢导致的查找失败
    '''
    import time
    try_count = 3
    for i in range(try_count):
        if i >= 3:
            break
        else:
            try:
                pid = _get_pid_by_port(port)
                return pid
            except RuntimeError:
                time.sleep(2)
            except Exception as err:
                raise err            
                
if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''
控件基类模块
'''
from __future__ import division
import six

from testbase.util import Timeout
from qt4c.mouse import Mouse, MouseFlag, MouseClickType
from qt4c.keyboard import Keyboard
from qt4c.util import MetisView
#__all__=['Control']
    
#===============================================================================
# 基础控件类定义
#===============================================================================
class Control(object):
    '''
    控件基类
    '''
    
    _timeout=Timeout(5,0.5) #控件查找timeout时间 
    def __init__(self):
        pass
    
    @property
    def Children(self):
        '''返回此控件的子控件。需要在子类中实现。
        '''
        raise NotImplementedError("请在%s类中实Children属性" % type(self))
    
#     @staticmethod
#     def exist(locator, root=None, timeout=10, interval=0.5):
#         '''在timeout时间内检查控件，返回找到的控件个数。
#         
#         :type locator: QPath
#         :param locator: 查找方式
#         :param root: 开始查找的父控件
#         :param timeout: 控件查找超时时间
#         :param interval: 查找间隔 
#         :return: 找到的控件个数
#         '''
#         
#         import util
#         ctrls = []
#         try: 
#             ctrls = util.Timeout(timeout, interval).retry(locator.search, (root,), (), lambda x: len(x) != 0)
#         except TimeoutError:
#             pass
#         return len(ctrls)
#     
#     @classmethod
#     def exist2(cls, *args, **kwargs):
#         '''在timeout时间内检查控件是否存在，返回True or False
#         
#         :param kwargs: 可变参数。代表控件类实例化所需的参数，比如说通过MainPanel类调用该函数，则参数传入qqapp实例
#         :param args: 可变参数。代表控件类实例化所需的参数，比如说通过MainPanel类调用该函数，则参数传入qqapp实例
#         :return: 该类型控件是否存在，True or False
#         :attention: 如果通过某个类的QPATH能找到多个符合条件的控件，会抛出ControlAmbiguousError，
#                                                     这表明该控件类的QPATH定义是不准确的
#         '''
#         return cls(*args, **kwargs).exist()
    
    def _click(self, mouseFlag, clickType, xOffset, yOffset):
        '''点击控件
        
        :type mouseFlag: qt4c.mouse.MouseFlag
        :param mouseFlag: 鼠标按钮
        :type clickType: qt4c.mouse.MouseClickType 
        :param clickType: 鼠标动作
        :type xOffset: int
        :param xOffset: 距离控件区域左上角的偏移。
                                                               默认值为None，代表控件区域x轴上的中点；
                                                               如果为负值，代表距离控件区域右边的绝对值偏移；
        :type yOffset: int
        :param yOffset: 距离控件区域左上角的偏移。
                                                               默认值为None，代表控件区域y轴上的中点；
                                                              如果为负值，代表距离控件区域上边的绝对值偏移；
        '''
        if not self.BoundingRect:
            return
        (l, t, r, b) = self.BoundingRect.All
        if xOffset is None:
            x = (l + r) // 2
        else:
            if xOffset < 0:
                x = r + xOffset
            else:
                x = l + xOffset
        if yOffset is None:
            y = (t + b) // 2
        else:
            if yOffset < 0:
                y = b + yOffset
            else:
                y = t + yOffset        
        Mouse.click(x, y, mouseFlag, clickType)
        
    def click(self, mouseFlag=MouseFlag.LeftButton, 
                    clickType=MouseClickType.SingleClick,
                    xOffset=None, yOffset=None):
        '''点击控件
        
        :type mouseFlag: qt4c.mouse.MouseFlag
        :param mouseFlag: 鼠标按钮
        :type clickType: qt4c.mouse.MouseClickType 
        :param clickType: 鼠标动作
        :type xOffset: int
        :param xOffset: 距离控件区域左上角的偏移。默认值为None，代表控件区域x轴上的中点。 如果为负值，代表距离控件区域右上角的x轴上的绝对值偏移。
        :type yOffset: int
        :param yOffset: 距离控件区域左上角的偏移。 默认值为None，代表控件区域y轴上的中点。如果为负值，代表距离控件区域右上角的y轴上的绝对值偏移。
        '''
        self.hover()
        x, y = self._getClickXY(xOffset, yOffset)
        Mouse.click(x, y, mouseFlag, clickType)
        
    def _getClickXY(self, xOffset, yOffset):
        '''通过指定的偏移值确定具体要点击的x,y坐标
        '''
        if not self.BoundingRect:
            return
        (l, t, r, b) = self.BoundingRect.All
        if xOffset is None:
            x = (l + r) // 2
        else:
            if xOffset < 0:
                x = r + xOffset
            else:
                x = l + xOffset
        if yOffset is None:
            y = (t + b) // 2
        else:
            if yOffset < 0:
                y = b + yOffset
            else:
                y = t + yOffset
        return x, y       
    
    def doubleClick(self, xOffset=None, yOffset=None):
        """左键双击，参数参考click函数
        """
        self.click(MouseFlag.LeftButton, MouseClickType.DoubleClick, xOffset, yOffset)
    
    def hover(self):
        """鼠标移动到该控件上
        """
        if not self.BoundingRect:
            return
#        (l, t, r, b) = (self.BoundingRect.Left, 
#                        self.BoundingRect.Top,
#                        self.BoundingRect.Right,
#                        self.BoundingRect.Bottom
#                        )

        x, y = self.BoundingRect.Center.All
        Mouse.move(x, y)
        
    def rightClick(self, xOffset=None, yOffset=None):
        """右键双击，参数参考click函数
        """
        self.click(MouseFlag.RightButton, xOffset=xOffset, yOffset=yOffset)
    
       
#    def leftClick(self):
#        self.click(MouseFlag.LeftButton)
        
    def drag(self, toX, toY):
        '''拖拽控件到指定位置
        '''
        if not self.BoundingRect:
            return
        (l, t, r, b) = (self.BoundingRect.Left, 
                        self.BoundingRect.Top,
                        self.BoundingRect.Right,
                        self.BoundingRect.Bottom
                        )
        x, y = (l + r) // 2, (t + b) // 2
        Mouse.drag(x, y, toX, toY)

    def scroll(self, backward=True):
        '''发送鼠标滚动命令
        '''
        self.hover()
        Mouse.scroll(backward)
    
    def sendKeys(self, keys):
        '''发送按键命令
        '''
        self.setFocus()
        Keyboard.inputKeys(keys)

    def setFocus(self):
        '''设控件为焦点
        '''
        raise NotImplementedError('please implement in sub class')
    
    def waitForValue(self, prop_name, prop_value, timeout=10, interval=0.5, regularMatch=False):
        """等待控件属性值出现, 如果属性为字符串类型，则使用正则匹配
        
        :param prop_name: 属性名字
        :param prop_value: 等待出现的属性值
        :param timeout: 超时秒数, 默认为10
        :param interval: 等待间隔，默认为0.5
        :param regularMatch: 参数 property_name和waited_value是否采用正则表达式的比较。默认为不采用（False）正则，而是采用恒等比较。
        """
        Timeout(timeout, interval).waitObjectProperty(self, prop_name, prop_value, regularMatch)
    
    @property
    def BoundingRect(self):
        """返回窗口大小。未实现！
        """
        raise NotImplementedError("please implement in sub class")
    
    def equal(self, other):
        '''判断两个对象是否相同。未实现!
        
        :type other: Control
        :param other: 本对象实例
        '''
        raise NotImplementedError("please implement in sub class")
        
    def __eq__(self, other):
        """重载对象恒等操作符(==)
        """
        return self.equal(other)
    
    def __ne__(self, other):
        """重载对象不等操作符(!=)
        """
        return (not self.equal(other))
    
    def get_metis_view(self):
        '''返回MetisView
        '''
        return MetisView(self)
    
class ControlContainer(object):
    '''控件集合接口
    
    当一个类继承本接口，并设置Locator属性后，该类可以使用Controls属于获取控件。如
    
    >>>class SysSettingWin(uia.UIAWindows, ControlContainer)
            def __init__(self):
                locators={'常规页': {'type':uia.Control, 'root':self, 'locator'='PageBasicGeneral'},
                          '退出程序单选框': {'type':uia.RadioButton, 'root':'@常规页','locator'=QPath("/name='ExitPrograme_RI' && maxdepth='10'")}}
                self.updateLocator(locators)
                                 
    则SysSettingWin().Controls['常规页']返回设置窗口上常规页的uia.Control实例,
    而SysSettingWin().Controls['退出程序单选框']，返回设置窗口的常规页下的退出程序单选框实例。
    其中'root'='@常规页'中的'@常规页'表示参数'root'的值不是这个字符串，而是key'常规页'指定的控件。
    '''
    
    def __init__(self):
        self._locators = {}
    
    def __findctrl_recur(self, ctrlkey):
        if not (ctrlkey in self._locators.keys()):
            raise NameError("%s没有名为'%s'的子控件！" % (type(self), ctrlkey))
        params = self._locators[ctrlkey].copy()
        ctrltype = params['type']
        del params['type']
        for key in params:
            value = params[key]
            if isinstance(value, six.string_types) and value.startswith('@'):
                params[key] = self.__findctrl_recur(value[1:])
        return ctrltype(**params)
        
        
    def __getitem__(self, index):
        '''获取index指定控件
        
        :type index: string
        :param index: 控件索引，如'查找按钮'  
        '''
        return self.__findctrl_recur(index)
    
    def clearLocator(self):
        '''清空控件定位参数
        '''
        self._locators = {}
    
    def hasControlKey(self, control_key):
        '''是否包含控件control_key
        
        :rtype: boolean
        '''
        return control_key in self._locators
    
    def updateLocator(self, locators):
        '''更新控件定位参数
        
        :type locators: dict
        :param locators: 定位参数，格式是 {'控件名':{'type':控件类, 控件类的参数dict列表}, ...}
        '''
        self._locators.update(locators)    
     
    def isChildCtrlExist(self, childctrlname):
        '''判断指定名字的子控件是否存在
        
        :param childctrlname: 指定的子控件名称
        :type childctrlname: str
        :rtype: boolean
        '''
        return self.Controls[childctrlname].exist()
    
    @property
    def Controls(self):
        '''返回控件集合。使用如foo.Controls['最小化按钮']的形式获取控件
        '''
        return self   
  
if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''浏览器类库
'''

class EnumBrowserType(object):
    IE="iexplorer"
    TT="ttraverler"
    FireFox="firefox"
    Chrome="chrome"
    QQBrowser="qqbrowser"

browser_list=[]
for name in dir(EnumBrowserType):
    if not name.startswith('__'):
        browser_list.append(getattr(EnumBrowserType,name))

def _get_default_browser1():
    '''获取xp下的默认浏览器
    '''
    import win32con, win32api
    hkey = win32con.HKEY_CLASSES_ROOT
    subkey = r'http\shell\open\command'
    hkey = win32api.RegOpenKey(hkey, subkey)
    brwcmd = win32api.RegQueryValue(hkey, None)
    win32api.RegCloseKey(hkey)
    for browser_type in browser_list:
        if brwcmd.lower().find(browser_type)>=0:
            return browser_type
    else:
        raise RuntimeError("未支持浏览器%s" % brwcmd)#Exception("浏览器%s还未支持" % brwchoice)

def _get_default_browser2():
    '''获取vista或以上的默认浏览器
    '''
    import win32con, win32api, pywintypes
    hkey = win32con.HKEY_CURRENT_USER
    subkey = r'Software\Microsoft\Windows\Shell\Associations\UrlAssociations\http\UserChoice'
    win32api.RegOpenCurrentUser()
    brwchoice = 'IE.HTTP'
    try:
        hkey = win32api.RegOpenKey(hkey, subkey)
        brwchoice = win32api.RegQueryValueEx(hkey, 'ProgId')[0]
        win32api.RegCloseKey(hkey)
    except pywintypes.error: # subkey没有设过，因此默认是IE
        return EnumBrowserType.IE
    brwmap = {'FirefoxURL': EnumBrowserType.FireFox, 
              'IE.HTTP': EnumBrowserType.IE, 
              'TTraveler.HTTP':EnumBrowserType.TT}
    if brwchoice in brwmap.keys():
        return brwmap[brwchoice]
    else:
        raise RuntimeError("未支持浏览器%s" % brwchoice)
    
def get_default_browser():
    '''获取默认浏览器类型
    
    :rtype: EnumBrowserType
    '''
    import sys
    winver = sys.getwindowsversion()
    if winver[0]<=5 : #xp or below
        return _get_default_browser1()
    elif winver[0] >=6: # vista or above
        return _get_default_browser2()

if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#

'''IWebWiew接口定义及Windows平台上的实现
'''

import sys, logging
fmt = logging.Formatter('%(asctime)s %(thread)d %(message)s')
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(fmt)
logging.root.addHandler(handler)



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''
Window的控件模块
'''

from __future__ import division
import locale
import win32api
import win32gui
import win32process
import win32con
import winerror
import commctrl
import copy
import ctypes
import os
import time
import types
import six

from testbase.util import LazyInit, Timeout

from qt4c.util import ProcessMem, Rectangle
from qt4c.mouse import Mouse, MouseFlag, MouseClickType
from qt4c.exceptions import ControlAmbiguousError, ControlNotFoundError, TimeoutError
from qt4c import accessible, control, wintypes, util
from qt4c.keyboard import Keyboard

class _CWindow(object):
    '''
    Win32 Window类（bridge）
    '''
    
    #TODO：将Control类中的Windows窗口接口移到此类
    
    def __init__(self, hwnd):
        '''构造函数
        :param hwnd: 窗口句柄
        :type hwnd: integer
        '''
        self._hwnd = hwnd 
        
    @property
    def HWnd(self):
        '''窗口句柄
        '''
        return self._hwnd
    
    
class Control(control.Control):
    '''
    Win32 Window类，实现Win32窗口常用属性。
    '''

    def __init__(self, root=None, locator=None):
        '''Constructor
        
        :type root: wincontrols.Control
        :param root: 开始查询的窗口。 如果是None，表示用desktop开始查找。
        :type locator: qpath.QPath
        :param locator: 查询类的实例, 如果是None，则将root作为此窗口。
        '''
        
        control.Control.__init__(self)
        if locator is None:
            if (isinstance(root, six.integer_types) or 
                isinstance(root, Control) or 
                isinstance(root, _CWindow) or
                root is None):
                pass
            else:
                raise TypeError("root不支持该类型%s" % type(root))
        self._root = root
        self._locator = locator
        self._wndobj = LazyInit(self, '_wndobj', self._init_wndobj )
        
    def _init_wndobj(self):
        '''初始化Win32窗口对象
        '''
        if self._locator is None and self._root is None:
            wndobj =  _CWindow(int(win32gui.GetDesktopWindow()))
        elif self._locator is None:
            if isinstance(self._root, six.integer_types): #root is a hwnd
                wndobj = _CWindow(self._root)
            elif isinstance(self._root, Control):
                wndobj = _CWindow(self._root.HWnd)
            elif isinstance(self._root, _CWindow):
                wndobj = self._root
        else:
            try:
                kwargs = {'root': self._root}
                foundctrls =  self._timeout.retry(self._locator.search, kwargs, (), lambda x: len(x)>0 )
            except TimeoutError as erro:
                raise ControlNotFoundError("<%s>中的%s查找超时：%s" % (self._locator, self._locator.getErrorPath(), erro))
            nctrl = len(foundctrls)
            if (nctrl > 1):
                raise ControlAmbiguousError("<%s>找到%d个控件" % (self._locator, nctrl))
            wndobj = _CWindow(foundctrls[0].HWnd)
        return wndobj
    
    @staticmethod
    def __enum_childwin_callback(hwnd, hwnds):
        parent = hwnds[0]
        if parent == None:
            hwnds.append(hwnd)
        else:
            hparent = ctypes.windll.user32.GetAncestor(hwnd, win32con.GA_PARENT)
            if hparent == parent:
                hwnds.append(hwnd)  
             
    def __validCtrlNum(self, ctrls):
        return (len(ctrls) > 0)
       
    def equal(self, other):
        '''判断两个对象是否相同。
        
        :type other: Control
        :param other: 本对象实例
        '''
        if not isinstance(other, Control):
            return False
        return (self.HWnd == other.HWnd)
    
    def exist(self):
        '''判断控件是否存在
        '''
        if self._locator is None and self._root is None:
            return True
        elif self._locator is None:
            if isinstance(self._root, six.integer_types): #root is a hwnd
                return True
            elif isinstance(self._root, Control):
                return self._root.exist()
            elif isinstance(self._root, _CWindow):
                return self._root.Valid
        else:
            if isinstance(self._root, Control):
                if not self._root.exist():
                    return False
            foundctrls = self._locator.search(root=self._root)
            nctrl = len(foundctrls)
            if (nctrl > 1):
                raise ControlAmbiguousError("<%s>找到%d个控件" % (self._locator, nctrl))
            if (nctrl < 1):
                return False
            else:
                return True    
    
    def wait_for_exist(self, timeout, interval ):
        '''等待控件存在
        '''
        Timeout(timeout, interval).retry(self.exist, (), (), lambda x:x==True)
        
    def waitForExist(self, timeout, interval ):
        '''等待控件存在
        '''
        Timeout(timeout, interval).retry(self.exist, (), (), lambda x:x==True)
        
    def wait_for_invalid(self, timeout=10.0, interval=0.5 ):
        '''等待控件失效
        '''
        Timeout(timeout, interval).retry(self.exist, (), (), lambda x:x==False)
        
    def waitForInvalid(self, timeout=10.0, interval=0.5 ):
        '''等待控件失效
        '''
        Timeout(timeout, interval).retry(self.exist, (), (), lambda x:x==False) 
        
    @property
    def BoundingRect(self):
        """返回窗口大小
        
        :rtype: util.Rectangle
        :return: util.Rectangle实例
        """
        return util.Rectangle(win32gui.GetWindowRect(self.HWnd))
    
    @property
    def Caption(self):
        """返回窗口标题
        
        :rtype: StringType
        :return: 窗口标题
        """
        buf_size = 0
        try:
            textlength = ctypes.c_long(0)
            hr = ctypes.windll.user32.SendMessageTimeoutA(self.HWnd, win32con.WM_GETTEXTLENGTH, 0, 0, 0, 200, ctypes.byref(textlength))
            if hr == 0 or textlength.value < 0:
                return ""
            buf_size = textlength.value + 1
        except win32gui.error:
            return ""

        if buf_size <= 0:
            return ""
        # pybuffer = win32gui.PyMakeBuffer(buf_size)
        # ret = win32gui.SendMessage(self.HWnd, win32con.WM_GETTEXT, buf_size, pybuffer)

        pybuffer = ctypes.create_unicode_buffer(buf_size)
        ret = ctypes.windll.user32.SendMessageW(self.HWnd, win32con.WM_GETTEXT, buf_size, ctypes.byref(pybuffer))
        if ret:
            text = pybuffer.value
            try:
                if six.PY2:
                    text = text.encode('utf8')
                return text
            except Exception:
                return text
        else:
            return ""

    @property
    def Children(self):
        """返回子控件列表
        
        :rtype: ListType
        """
        hwnds = []
        if self.HWnd == win32gui.GetDesktopWindow():
            hwnds.append(None)
            win32gui.EnumWindows(self.__enum_childwin_callback, hwnds)
        else:
            hwnds.append(self.HWnd)
            try:
                win32gui.EnumChildWindows(self.HWnd, self.__enum_childwin_callback, hwnds)
            except win32gui.error as e:
                # if e[0]== 0 or e[0]==1400: #1400是无效窗口错误
                if e.winerror == 0 or e.winerror == winerror.ERROR_INVALID_WINDOW_HANDLE: #1400是无效窗口错误
                    pass
                else:
                    raise e
            
        del hwnds[0]
        return [Window(root=hwnd) for hwnd in hwnds]

    @property
    def ClassName(self):
        '''返回窗口类名
        '''
        text = win32gui.GetClassName(self.HWnd)
        os_encoding = locale.getdefaultlocale(None)[1]
        try:
            if six.PY2:
                text = text.decode(os_encoding)
            return text
        except UnicodeDecodeError:
            return text
    
    @property
    def ControlId(self):
        '''返回控件ID
        '''
        return win32gui.GetDlgCtrlID(self.HWnd)
    
    @property
    def Enabled(self):
        '''此控件是否可用
        '''
        bEnable = win32gui.IsWindowEnabled(self.HWnd)
        if bEnable == 1:
            return True
        else:
            return False
    
    @property
    def ExStyle(self):
        """此控件的扩展样式
        """
        return win32gui.GetWindowLong(self.HWnd, win32con.GWL_EXSTYLE)
        
    @property
    def HWnd(self):
        return self.hwnd

    @property
    def hwnd(self):
        return self._wndobj.HWnd
    
    @property        
    def Parent(self):
        """返回父窗口
        
        :rtype: Window
        :return: 获取父窗口
        :attention: 如果是desktop窗口，则返回None;如果是顶级窗口，则返回desktop窗口;否则返回父窗口 。
        """
        hwind_desktop = win32gui.GetDesktopWindow()
        if self.HWnd == hwind_desktop:
            return None

        isChild = ((self.Style & win32con.WS_CHILDWINDOW) > 0)
        if isChild:
            hwnd = win32gui.GetParent(self.HWnd)
            return Window(root=hwnd)
        else:
            return Window(root=hwind_desktop)
        
    @property
    def ProcessId(self):
        pid = win32process.GetWindowThreadProcessId(self.HWnd)[1]
        return pid
    
    @property
    def Style(self):
        """此控件的样式
        """
        return win32gui.GetWindowLong(self.HWnd, win32con.GWL_STYLE)
        
    @property
    def Text(self):
        return self.Caption
    
    @Text.setter
    def Text(self, text):
        """设置文本
        
        :type text: StringType
        :param text: 设置的文本 
        """
        os_encoding = locale.getdefaultlocale(None)[1]
        if six.PY2:
            win32gui.SendMessage(self.HWnd, win32con.WM_SETTEXT, 0, text.decode('utf8').encode(os_encoding))
        else:
            win32gui.SendMessage(self.HWnd, win32con.WM_SETTEXT, 0, text)
        
    @property
    def ThreadId(self):
        '''窗口线程ID
        '''
        tid = win32process.GetWindowThreadProcessId(self.HWnd)[0]
        return tid
    
    @property
    def TopLevelWindow(self):
        '''返回控件的最上层窗口
        
        :rtype: Window
        '''
        topwnd = self
        while (topwnd.Style & win32con.WS_CHILDWINDOW) > 0:
            topwnd = topwnd.Parent
        return Window(root=topwnd)

    @property    
    def Valid(self):
        '''窗口是否存在
        '''
        if 0 == win32gui.IsWindow(self.HWnd):
            return False
        else:
            return True
    
    @property
    def Visible(self):
        '''此控件是否可见
        '''
        if not self.Valid:
            return False
        bVisible = win32gui.IsWindowVisible(self.HWnd)
        if bVisible == 1:
            return True
        else:
            return False
        
    @property
    def AccessibleObject(self):
        """返回AccessibleObject
        
        :rtype: qt4c.accessible.AccessibleObject
        """
        return accessible.AccessibleObject(self.HWnd)
        
    @property
    def Width(self):
        """宽度
        """
        if self.BoundingRect:
            return self.BoundingRect.Right - self.BoundingRect.Left
        return 0
    
    @property
    def Height(self):
        """高度
        """
        if self.BoundingRect:
            return self.BoundingRect.Bottom - self.BoundingRect.Top
        return 0
        
    def click(self, mouseFlag=MouseFlag.LeftButton, 
                    clickType=MouseClickType.SingleClick,
                    xOffset=None, yOffset=None):
        '''点击控件
        
        :type mouseFlag: qt4c.mouse.MouseFlag
        :param mouseFlag: 鼠标按钮
        :type clickType: qt4c.mouse.MouseClickType 
        :param clickType: 鼠标动作
        :type xOffset: int
        :param xOffset: 距离控件区域左上角的偏移。
                                                               默认值为None，代表控件区域x轴上的中点；
                                                               如果为负值，代表距离控件区域右上角的x轴上的绝对值偏移；
        :type yOffset: int
        :param yOffset: 距离控件区域左上角的偏移。
                                                               默认值为None，代表控件区域y轴上的中点；
                                                              如果为负值，代表距离控件区域右上角的y轴上的绝对值偏移；
        '''
        x, y = self._getClickXY(xOffset, yOffset)
        Mouse.sendClick(self.HWnd, x,y, mouseFlag, clickType)
        
    def setFocus(self):
        '''将此控件设为焦点
        '''
#        self.bringForeground() #comment out because of bug 4973401
        current_id = win32api.GetCurrentThreadId()
        target_id = self.ThreadId
        win32process.AttachThreadInput(target_id,current_id,True)
        win32gui.SetFocus(self.HWnd)
        win32process.AttachThreadInput(target_id,current_id,False)

class ListViewItem(control.Control):
    """sysListView32的每一项
    """
    def __init__(self, parent, item_index):
        self._parent = parent
        self.item_index = item_index
        control.Control.__init__(self)
    
    @property
    def SubItems(self):
        pass
    
    @property
    def BoundingRect(self):
        """获取ListView的某项Item的文本
        """
        rect = wintypes.RECT()
        rcsize = ctypes.sizeof(wintypes.RECT)
        pm = util.ProcessMem(self._parent.ProcessId, buffer_size=rcsize)
        rect.left = commctrl.LVIR_SELECTBOUNDS
        pm.write(ctypes.byref(rect), rcsize)
        
        # Fill in the requested item
        retval = win32gui.SendMessage(self._parent.HWnd,
                                      commctrl.LVM_GETITEMRECT,
                                      self.item_index,
                                      pm.Buffer)
    
        # if it succeeded
        if not retval:
            raise RuntimeError("Did not succeed in getting rectangle")
        pm.read(ctypes.byref(rect), rcsize)
        parentRC = self._parent.BoundingRect
        
        return util.Rectangle((parentRC.Left+rect.left,
                               parentRC.Top+rect.top,
                               parentRC.Left+rect.right,
                               parentRC.Top+rect.bottom))
    
    @property
    def Text(self):
        """获取ListView的某项Item的文本
        """
        dwPid = win32process.GetWindowThreadProcessId(self._parent.HWnd)[1]
        hProcess = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, False, dwPid).Detach()
        
        if 'PROGRAMFILES(X86)' in os.environ:
            lvi = wintypes.LVITEM64()
        else:
            lvi = wintypes.LVITEM()
        pTextBuf = ctypes.c_char_p(six.b('\0') * win32con.MAX_PATH)
    
        plvi = ctypes.windll.kernel32.VirtualAllocEx(hProcess, None, ctypes.sizeof(wintypes.LVITEM), win32con.MEM_COMMIT, win32con.PAGE_READWRITE)  # @UndefinedVariable
        pBuf = ctypes.windll.kernel32.VirtualAllocEx(hProcess, None, win32con.MAX_PATH, win32con.MEM_COMMIT, win32con.PAGE_READWRITE) # @UndefinedVariable
        lvi.iSubItem = 0
        lvi.pszText = pBuf
        lvi.cchTextMax = win32con.MAX_PATH
        lvi.mask = commctrl.LVIF_TEXT
        ctypes.windll.kernel32.WriteProcessMemory(hProcess, plvi, ctypes.byref(lvi), ctypes.sizeof(wintypes.LVITEM), None) # @UndefinedVariable
        ctypes.windll.kernel32.WriteProcessMemory(hProcess, pBuf, 0, win32con.MAX_PATH, None) # @UndefinedVariable
        win32gui.SendMessage(self._parent.HWnd, commctrl.LVM_GETITEMTEXT, self.item_index, plvi)
        ctypes.windll.kernel32.ReadProcessMemory(hProcess, pBuf, pTextBuf, win32con.MAX_PATH, None) # @UndefinedVariable
        ctypes.windll.kernel32.VirtualFreeEx(hProcess, plvi, ctypes.sizeof(wintypes.LVITEM), win32con.MEM_DECOMMIT) # @UndefinedVariable
        ctypes.windll.kernel32.VirtualFreeEx(hProcess, pBuf, win32con.MAX_PATH, win32con.MEM_DECOMMIT) # @UndefinedVariable
        win32api.CloseHandle(hProcess)
        pTextBuf = pTextBuf.value.decode('gbk').encode('utf8')
        return pTextBuf
    
class ListView(Control):
    """sysLisView32 控件类型
    """
    @property
    def ItemCount(self):
        "The number of items in the ListView"
        return win32gui.SendMessage(self.HWnd,commctrl.LVM_GETITEMCOUNT)

    @property
    def Items(self):
        return self
    
    def __iter__(self):
        cnt = self.ItemCount
        for i in range(cnt):
            yield ListViewItem(self, i)  
            
    def __getitem__(self, key):
        if isinstance(key, int):
            cnt = self.ItemCount
            if key >= cnt:
                raise IndexError("key超出下标范围!")
            return ListViewItem(self, key)
        elif isinstance(key, six.string_types):
            if(isinstance(key, six.text_type)):
                key = key.encode('utf8')
            for item in self:
                if item.Text == key:
                    return item
            raise ValueError("cannot find Listview item of text %s" %key)  
        raise TypeError('参数key=%s, 不是int或string' % key)    

class TextBox(Control):
    """编辑控件
    """
    pass

class Window(Control, control.ControlContainer):
    '''
    Win32 Window类，实现Win32窗口常用属性。
    '''
    def __init__(self, root=None, locator=None):
        '''Constructor
        
        :type locator: 类实例 
        :param locator: 查询类的实例, 如果是None，则将root作为此窗口。
        :type root: Window或者window句柄 
        :param root: 开始查询的窗口。 如果是None，表示用desktop开始查找。
        '''
        Control.__init__(self, root=root, locator=locator)
        control.ControlContainer.__init__(self)  
        
    @property
    def Maximized(self):
        """该窗口是否最大化
        """
        if not self.Valid:
            raise RuntimeError("句柄无效，窗口已销毁!")
        if((self.Style & win32con.WS_MAXIMIZE) > 0):
            return True
        else:
            return False
    
    @property
    def Minimized(self):
        """该窗口是否最小化
        """
        if not self.Valid:
            raise RuntimeError("句柄无效，窗口已销毁!")
        return win32gui.IsIconic(self.HWnd)
    
    @property
    def OwnerWindow(self):
        '''此窗口的所有者窗口
        
        :rtype: Window
        :return: 获取Owner Window
        '''
        if not self.Valid:
            return None
        hwnd = win32gui.GetWindow(self.HWnd, win32con.GW_OWNER)
        if hwnd == None:
            return None
        else:
            return Window(root=hwnd)
    
    @property
    def PopupWindow(self):
        '''此窗口的弹出窗口
        
        :rtype: Window
        :return: 获取EnabledPopup
        '''
        hwnd = win32gui.GetWindow(self.HWnd, win32con.GW_ENABLEDPOPUP)
        if hwnd == 0:
            return None
        else:
            return Window(root=hwnd)
        
    @property
    def TopMost(self):
        """是否具有总在最前端
        """
        if((self.ExStyle & win32con.WS_EX_TOPMOST) > 0):
            return True
        else:
            return False
        
    def bringForeground(self):
        '''将窗口设为最前端窗口
        '''
        fwnd = win32gui.GetForegroundWindow()
        if(fwnd == self.HWnd):
            return
        if fwnd == 0:
            try:
                win32gui.SetForegroundWindow(self.HWnd)
                return
            except win32api.error:
                # 防止有菜单弹出导致异常
                Keyboard.inputKeys('{ESC}')
                win32gui.SetForegroundWindow(self.HWnd)
                return
        
        ftid, _ = win32process.GetWindowThreadProcessId(fwnd)
        wtid, _ = win32process.GetWindowThreadProcessId(self.HWnd)
                
        ctypes.windll.user32.AttachThreadInput(wtid, ftid, True)
#        if ret == 0:
#            raise Exception("AttachThreadInput(attach) with exception, error code:%s" % win32api.GetLastError())

        try:
            util.Timeout(5, 0.5).retry(  ctypes.windll.user32.SetForegroundWindow, 
                                         (self.HWnd, ), 
                                         (), 
                                         lambda x: win32gui.GetForegroundWindow() == self.HWnd)
        except TimeoutError:
            pass
        
        ctypes.windll.user32.AttachThreadInput(wtid, ftid, False)
#        if ret == 0:
#            raise Exception("AttachThreadInput(detach) with exception, error code:%s" % win32api.GetLastError())
            
    
    def _wait_for_disabled_or_invisible(self, timeout=60, interval=0.5):
        tt = util.Timeout(timeout, interval)
        tt.retry(lambda x: x, (self, ), (), lambda x: x.Enabled==False or x.Valid==False or x.Visible==False)
        
    def close(self):
        '''关闭窗口
        
        :rtype: bool
        :return: 窗口销毁返回True，否则返回False
        '''
        win32api.PostMessage(self.HWnd,win32con.WM_CLOSE,0,0)
        self._wait_for_disabled_or_invisible() #有弹出窗口时，此窗口会变为disabled
        if self.Valid == True:
            return False
        else:
            return True 
        
    
    def hide(self):
        '''隐藏
        '''
        win32gui.ShowWindow(self.HWnd, win32con.SW_HIDE)
    
    def maximize(self):
        '''最大化窗口
        '''
        win32api.PostMessage(self.HWnd,win32con.WM_SYSCOMMAND,win32con.SC_MAXIMIZE,0) 
    
    def minimize(self):
        '''最小化窗口
        '''
        win32api.SendMessage(self.HWnd,win32con.WM_SYSCOMMAND,win32con.SC_MINIMIZE,0) 
    
    def resize(self, width, height):
        '''设置窗口大小
        '''
        x, y = self.BoundingRect.Left, self.BoundingRect.Top
        win32gui.MoveWindow(self.HWnd, x, y, width, height, True)
        
    def restore(self):
        '''恢复窗口
        '''
        win32gui.ShowWindow(self.HWnd,win32con.SW_RESTORE)

    def show(self):
        '''显示窗口
        '''
        win32gui.ShowWindow(self.HWnd, win32con.SW_SHOW)
        
    def move(self, x, y):
        """移动窗口
        """
        win32gui.MoveWindow(self.HWnd, x, y, self.Width, self.Height, True)
        
    def waitForInvalid(self, timeout=10.0, interval=0.5):
        '''等待窗口失效
        
        :type timeout: float 
        :param timeout: 超时秒数
        '''
        if False == self.Valid:
            return 
        self.waitForValue('Valid', False, timeout, interval, False)
        
    def waitForInvisible(self, timeout=10.0, interval=0.5):
        '''等待窗口消失
        
        :type timeout: float 
        :param timeout: 超时秒数
        '''
        if False == self.Visible:
            return 
        self.waitForValue('Visible', False, timeout, interval, False)

#class ShellTrayWnd(Control):
#    """Tray Window"""
#    class ABE(object):
#        """Windows Taskbar的停靠位置"""
#        LEFT,TOP,RIGHT,BOTTOM = range(4)
#        
#    def __init__(self):
#        import qpath
#        qp = qpath.QPath("/ClassName='Shell_TrayWnd'")
#        Control.__init__(self, locator=qp)
#        
#    @property
#    def NotifyBar(self):
#        return TrayNotifyBar()
#    
#    @property
#    def TaskBar(self):
#        return TrayTaskBar()
#    
#    @property
#    def Locked(self):
#        """任务栏是否被锁定"""
#        uEdge = self.Edge
#        if uEdge == ShellTrayWnd.ABE.BOTTOM or uEdge == ShellTrayWnd.ABE.TOP:
#            oLeft = self.TaskBar.BoundingRect.Left
#            win32gui.SendMessage(self.HWnd, win32con.WM_COMMAND, 424, 0)
#            nLeft = self.TaskBar.BoundingRect.Left
#            win32gui.SendMessage(self.HWnd, win32con.WM_COMMAND, 424, 0)
#            if nLeft > oLeft:
#                return True
#            else:
#                return False
#        else:
#            oTop = self.TaskBar.BoundingRect.Top
#            win32gui.SendMessage(self.HWnd, win32con.WM_COMMAND, 424, 0)
#            nTop = self.TaskBar.BoundingRect.Top
#            win32gui.SendMessage(self.HWnd, win32con.WM_COMMAND, 424, 0)
#            if nTop > oTop:
#                return True
#            else:
#                return False
#        
#    @Locked.setter
#    def Locked(self, lock):
#        """锁定/解锁任务栏
#        @param lock: bool
#        @param lock: 是否锁定任务栏  
#        """
#        if self.Locked != lock:
#            win32gui.SendMessage(self.HWnd, win32con.WM_COMMAND, 424, 0)
#    
#    @property
#    def Edge(self):
#        """返回Windows Taskbar的停靠位置
#        @rtype: ShellTrayWnd.ABE
#        """
#        abd = wintypes.APPBARDATA()
#        abd.cbSize = ctypes.sizeof(abd)
#        abd.hWnd = self.HWnd
#        ctypes.windll.shell32.SHAppBarMessage(5, ctypes.byref(abd))
#        return abd.uEdge
#    
#    def setPosition(self, edge):
#        """设置Windows Taskbar的停靠位置
#        @param edge: ShellTrayWnd.ABE
#        """
#        from mouse import Mouse
#        import time
#        fromX = self.TaskBar.BoundingRect.Right
#        fromY = self.TaskBar.BoundingRect.Bottom
#        
#        uEdge = self.Edge
#        if uEdge == ShellTrayWnd.ABE.LEFT:
#            fromX = fromX - 1 
#        elif uEdge == ShellTrayWnd.ABE.TOP:
#            fromY = fromY - 1
#        
#        pDevMode = win32api.EnumDisplaySettings(None, win32con.ENUM_CURRENT_SETTINGS)
#        toX = pDevMode.PelsWidth / 2
#        toY = pDevMode.PelsHeight / 2
#        if edge == ShellTrayWnd.ABE.LEFT:
#            toX = 5
#        elif edge == ShellTrayWnd.ABE.TOP:
#            toY = 5
#        elif edge == ShellTrayWnd.ABE.RIGHT:
#            toX = pDevMode.PelsWidth - 5
#        else:
#            toY = pDevMode.PelsHeight - 5
#        
#        Mouse.drag(fromX, fromY, toX, toY)
#        time.sleep(3)
    
class TrayNotifyBar(Control):
    """系统的通知区域
    """
    def __init__(self):
        from qt4c import qpath
        import platform
        if platform.release() == 'XP':
            qp = qpath.QPath("/ClassName='Shell_TrayWnd'/ClassName='TrayNotifyWnd'/ClassName='ToolbarWindow32' && MaxDepth='5' && Instance='0'")
        else:
            qp = qpath.QPath("/ClassName='Shell_TrayWnd'/ClassName='TrayNotifyWnd'/ClassName='SysPager'/ClassName='ToolbarWindow32' && MaxDepth='5'")
        Control.__init__(self, locator=qp)
    
    @property
    def Items(self):
        """返回TrayNotifyBar的全部TrayNotifyIcon
        """
        itemcount = win32gui.SendMessage(self.HWnd, commctrl.TB_BUTTONCOUNT, 0, 0)
        tb = wintypes.TBBUTTON()
        tbsize = ctypes.sizeof(wintypes.TBBUTTON)
        td = wintypes.TRAYDATA()
        tdsize = ctypes.sizeof(wintypes.TRAYDATA)
        pm = util.ProcessMem(self.ProcessId, buffer_size=tbsize)
        items = []
        for i in range(itemcount):
            win32gui.SendMessage(self.HWnd, commctrl.TB_GETBUTTON, i, pm.Buffer)
            pm.read(ctypes.byref(tb), tbsize)
            pmtmp = util.ProcessMem(self.ProcessId, remote_buffer=tb.dwData)
            pmtmp.read(ctypes.byref(td), tdsize)
            
            isSeparator = tb.fsStyle & commctrl.TBSTYLE_SEP
            isEnable = tb.fsState & commctrl.TBSTATE_ENABLED
            if not isSeparator and isEnable:
                items.append(_TrayIcon(tb, td, self))
        
        return items
    
    def refresh(self):
        """刷新通知区域
        """
        width = self.BoundingRect.Width
        height = self.BoundingRect.Height
        hwnd = self.HWnd
        row = height // 16
        col = width // 16
        itemCnt = row * col
        for _ in range(itemCnt):
            l,t,r,b = self.BoundingRect.All
            for y in range(t,b,8):
                for x in range(l,r,8):
                    Mouse.postMove(hwnd, x, y)
#                    win32api.SetCursorPos((x,y))
            time.sleep(0.05)
    
    def __getitem__(self, key):
        if isinstance(key, six.string_types):
            return None
        else:
            for icon in self.Items:
                if icon.ProcessId == key:
                    return icon
            return None
    
class TrayTaskBar(Control):
    """系统的任务栏区域(win7以上不可用)
    """
    def __init__(self):
        from qt4c import qpath
        qp = qpath.QPath("/ClassName='Shell_TrayWnd'/ClassName='ReBarWindow32'/Text='运行应用程序' && ClassName='ToolbarWindow32' && MaxDepth='2'")
        Control.__init__(self, locator=qp)
    
    @property
    def Items(self):
        """返回TrayTaskBar的全部TrayNotifyIcon
        """
        itemcount = win32gui.SendMessage(self.HWnd, commctrl.TB_BUTTONCOUNT, 0, 0)
        tb = wintypes.TBBUTTON()
        tbsize = ctypes.sizeof(wintypes.TBBUTTON)
        td = wintypes.TRAYDATA()
        tdsize = ctypes.sizeof(wintypes.TRAYDATA)
        pm = util.ProcessMem(self.ProcessId, buffer_size=tbsize)
        items = []
        for i in range(itemcount):
            win32gui.SendMessage(self.HWnd, commctrl.TB_GETBUTTON, i, pm.Buffer)
            pm.read(ctypes.byref(tb), tbsize)
            pmtmp = util.ProcessMem(self.ProcessId, remote_buffer=tb.dwData)
            pmtmp.read(ctypes.byref(td), tdsize)
            isSeparator = tb.fsStyle & commctrl.TBSTYLE_SEP
            isEnable = tb.fsState & commctrl.TBSTATE_ENABLED
            if not isSeparator and isEnable:
                items.append(_TrayIcon(tb, td, self))
        
        return items
    
    def __getitem__(self, key):
        if isinstance(key, six.string_types):
            for icon in self.Items:
                if icon.Tips == key:
                    return icon
            return None
        else:
            for icon in self.Items:
                if icon.ProcessId == key:
                    return icon
            return None
        
class _TrayIcon(control.Control):
    """通知栏或任务栏的项
    """
    def __init__(self, tbButton, trayData, notifyBar):
        '''不要直接实例化这个类，应该用TrayNotifyBar.Items得到这个类的实例
        '''
        self._tb = copy.deepcopy(tbButton)
        self._td = copy.deepcopy(trayData)
        self._notifybar = notifyBar

    @property
    def BoundingRect(self):
        """托盘图标的位置
        
        :rtype: util.Rectangle
        :return: util.Rectangle
        """
        barBoundingRect = self._notifybar.BoundingRect
        rcsize = ctypes.sizeof(wintypes.RECT)
        pm = util.ProcessMem(self._notifybar.ProcessId, buffer_size=rcsize)
        rc = wintypes.RECT()
        
        win32gui.SendMessage(self._notifybar.HWnd, 
                             commctrl.TB_GETRECT, 
                             self._tb.idCommand, 
                             pm.Buffer)
        pm.read(ctypes.byref(rc), rcsize)
        left = barBoundingRect.Left + rc.left
        top = barBoundingRect.Top + rc.top
        right = left + rc.right - rc.left
        bottom = top + rc.bottom - rc.top
        return util.Rectangle((left, top, right, bottom))
    
    @property
    def ProcessId(self):
        """图标所代表的进程Id
        """
        return win32process.GetWindowThreadProcessId(self._td.hwnd)[1]
        
    @property
    def State(self):
        """图标状态，隐藏or显示
        """
        return self._tb.fsState
    
    @property
    def Style(self):
        """图标风格，分隔符or自定义图片之类
        """
        return self._tb.fsStyle
    
    @property
    def Tips(self):
        """图标提示
        """
        pm = util.ProcessMem(self._notifybar.ProcessId, remote_buffer=self._tb.iString)
        # tips = ctypes.c_wchar_p('\0' * win32con.MAX_PATH)
        tips = ctypes.create_unicode_buffer(win32con.MAX_PATH)
        pm.read(tips, win32con.MAX_PATH)
        from qt4c.util import myEncode
        if isinstance(tips.value, six.text_type):
            return myEncode(tips.value, 'utf-8', 'UNICODE')
        else:
            return myEncode(tips.value, 'utf-8', 'GBK')
        
    @property
    def Visible(self):
        """是否可见
        """
        isVisible = not (self._tb.fsState & commctrl.TBSTATE_HIDDEN)
        return isVisible
    
    def click(self, mouseFlag=MouseFlag.LeftButton, 
                    clickType=MouseClickType.SingleClick,
                    xOffset=None, yOffset=None):
        '''点击控件 (同步操作)
        
        :type mouseFlag: qt4c.mouse.MouseFlag
        :param mouseFlag: 鼠标按钮
        :type clickType: qt4c.mouse.MouseClickType 
        :param clickType: 鼠标动作
        :type xOffset: int
        :param xOffset: 距离控件区域左上角的偏移。
                                                               默认值为None，代表控件区域x轴上的中点；
                                                               如果为负值，代表距离控件区域右上角的x轴上的绝对值偏移；
        :type yOffset: int
        :param yOffset: 距离控件区域左上角的偏移。
                                                               默认值为None，代表控件区域y轴上的中点；
                                                              如果为负值，代表距离控件区域右上角的y轴上的绝对值偏移；
        '''
        self.hover()
        x, y = self._getClickXY(xOffset, yOffset)
        Mouse.sendClick(self._notifybar.HWnd, x,y, mouseFlag, clickType)
    
    def destroy(self):
        """销毁该图标使之不再显示
        """
        win32gui.Shell_NotifyIcon(win32gui.NIM_DELETE, (self._td.hwnd, self._td.uID))
        
    def equal(self, other):
        if isinstance(other, _TrayIcon):
            return self.ProcessId==other.ProcessId
        return id(self)==other
 
 
class ComboBox(Control):
    """ComboBox 控件类型
    """
    @property 
    def Count(self):
        '''返回ComboBox的项目数
        '''
        return win32gui.SendMessage(self.HWnd,win32con.CB_GETCOUNT)
               
    @property
    def SelectedIndex(self):
        '''返回当选选中的索引值
        '''
        return win32gui.SendMessage(self.HWnd,win32con.CB_GETCURSEL)

    @SelectedIndex.setter
    def SelectedIndex(self, _item):
        """
        
        :desc: 选择ComboBox的某个item
        :param _item       in, 可以为数字索引值，也可以为文本
        """
#        if not CheckValidHWnd(self.HWnd):
#            return None
        
        itemCount = self.Count
        win32gui.SendMessage(self.HWnd, win32con.CB_SHOWDROPDOWN , True, 0)
        hComboLBoxWnd = win32gui.FindWindow('ComboLBox',None)
        ctrlId = win32gui.GetDlgCtrlID(hComboLBoxWnd)
        
        if isinstance(_item, six.integer_types):
            if((itemCount == win32con.CB_ERR) or
                            (_item >= itemCount) or (_item < -1)):
                raise IndexError("%d超出索引！" % _item)#Exception('Item is disability')
            win32gui.SendMessage(self.HWnd, win32con.CB_SETCURSEL, _item, 0)
            win32gui.PostMessage(self.HWnd, win32con.WM_COMMAND, (win32con.CBN_SELCHANGE << 16) | ctrlId, hComboLBoxWnd)
        else:
            item = _item
            itemList = []
            for i in range(itemCount):
                # pBuf = win32gui.PyMakeBuffer(win32con.MAX_PATH)
                # nchars = win32gui.SendMessage(self.HWnd, win32con.CB_GETLBTEXT, i, pBuf)
                # str_val = pBuf[:nchars]
                # itemList.append(str_val.decode('gbk').encode('utf8'))

                pBuf = ctypes.create_unicode_buffer(win32con.MAX_PATH)
                nchars = ctypes.windll.user32.SendMessageW(self.HWnd, win32con.CB_GETLBTEXT, i, ctypes.byref(pBuf))
                if nchars:
                    str_val = pBuf.value
                    if six.PY2:
                        str_val = str_val.encode('utf8')
                    itemList.append(str_val)
                else:
                    raise ValueError("未找到%s" % item) # Exception('Item is disability')
            if item not in itemList:
                raise ValueError("未找到%s" % item) # Exception('Item is disability')            
            win32gui.SendMessage(self.HWnd, win32con.CB_SETCURSEL, itemList.index(item), 0)
            win32gui.PostMessage(self.HWnd, win32con.WM_COMMAND, (win32con.CBN_SELCHANGE << 16) | ctrlId, hComboLBoxWnd)

    def _getTextByIndex(self, idx=-1):
        if idx < 0 or idx > self.Count - 1:
            return None
        
        # pBuf = win32gui.PyMakeBuffer(win32con.MAX_PATH)
        # nchars = win32gui.SendMessage(self.HWnd, win32con.CB_GETLBTEXT, idx, pBuf)
        # str_val = pBuf[:nchars]
        pBuf = ctypes.create_unicode_buffer(win32con.MAX_PATH)
        nchars = ctypes.windll.user32.SendMessageW(self.HWnd, win32con.CB_GETLBTEXT, idx, ctypes.byref(pBuf))
        if nchars:
            str_val = pBuf.value
            if six.PY2:
                str_val = str_val.encode('utf8')
        else:
            raise ValueError('获取文本失败')
        return str_val
        
    def getFullPath(self):
        retList = []
        curIdx = self.SelectedIndex
        itemTxt = self.Text
        
        if not itemTxt:
            return None
        retList.insert(0, itemTxt)
        for i in range(curIdx-1, -1, -1):
            itemTxt = self._getTextByIndex(i)
            if itemTxt and itemTxt.find(':') > -1:
                retList.insert(0, itemTxt[-3:-1])
                break
            else:
                retList.insert(0, itemTxt)
        bRoot = os.path.exists(retList[0] + '\\')
        bNotRoot = os.path.exists('\\'.join(retList))
        length = len(retList)
        if (length==1 and bRoot) or (length>1 and bNotRoot):
            return '\\'.join(retList)
        elif retList:
            return retList[-1]
        else:
            return None
        
class TreeView(Control):
    """Sys TreeView 32 控件类型
    """
    
    def count(self):
        '''返回TreeView的item数
        '''
        return win32gui.SendMessage(self.HWnd,commctrl.TVM_GETCOUNT)
        
    @property
    def Items(self):
        '''返回TreeView的首层节点列表
        
        :rtype: list
        '''
        itemlist = _ITEMLIST()
        #get root item first
        item = TreeViewItem(self.HWnd, win32gui.SendMessage(self.HWnd, commctrl.TVM_GETNEXTITEM, commctrl.TVGN_ROOT, 0))
        itemlist.append(item)
        #then get sibling of root item
        while True:
            try:
                item = item.NextSibling
            except ControlNotFoundError:
                break
            if item.BoundingRect.Height == 0:
                for subitem in item.Items:
                    itemlist.append(subitem)
            else:
                itemlist.append(item)
        return itemlist


class TreeViewItem(control.Control):
    """TreeView Item类，不要直接实例化这个类，而通过TreeView.Items来获取。
    """
    def __init__(self, hwnd, item):
        '''Constructor
        
        :param hwnd: TreeView的窗口句柄
        :param item: TreeView的内部控件对象
        '''    
        if item == 0:
            raise ControlNotFoundError("TreeView Item is not found or unavailable!")
        self._hwnd = hwnd
        self._pid = win32process.GetWindowThreadProcessId(hwnd)[1]
        self.__item = item
        control.Control.__init__(self)
    
    @property
    def HWnd(self):
        '''窗口句柄
        '''
        return self.hwnd

    @property
    def hwnd(self):
        '''窗口句柄
        '''
        return self._hwnd
    
    def ensureVisible(self):
        '''先保证item完全可见
        '''
        win32gui.SendMessage(self.HWnd, commctrl.TVM_ENSUREVISIBLE, 0, self.__item)
        
    def select(self):
        '''选中自身结点
        '''
        self.ensureVisible()
        win32gui.SendMessage(self.HWnd, commctrl.TVM_SELECTITEM, commctrl.TVGN_CARET, self.__item)
        
        
    @property
    def BoundingRect(self):              
        rect_size = ctypes.sizeof(wintypes.RECT)
        pmrect = ProcessMem(processId=self._pid, buffer_size=rect_size)
        pmrect.write(ctypes.byref(ctypes.c_int(self.__item)), rect_size)
        win32gui.SendMessage(self.HWnd, commctrl.TVM_GETITEMRECT, True, pmrect.Buffer)
        rect = wintypes.RECT()
        pmrect.read(ctypes.byref(rect), rect_size)
        tvrect = win32gui.GetWindowRect(self.HWnd)
        return Rectangle((tvrect[0]+rect.left, tvrect[1]+rect.top, tvrect[0]+rect.right, tvrect[1]+rect.bottom))

    @property
    def Items(self):        
        children_list = _ITEMLIST()
        child = win32gui.SendMessage(self.HWnd, commctrl.TVM_GETNEXTITEM, commctrl.TVGN_CHILD, self.__item)     
        while child:
            children_list.append(TreeViewItem(self.HWnd, child))
            child = win32gui.SendMessage(self.HWnd, commctrl.TVM_GETNEXTITEM, commctrl.TVGN_NEXT, child)            
        return children_list
    
    @property
    def Selected(self):        
        TVM_GETITEMSTATE = commctrl.TV_FIRST+39
        state = win32gui.SendMessage(self.HWnd, TVM_GETITEMSTATE, self.__item, commctrl.TVIS_SELECTED)
        return (state & commctrl.TVIS_SELECTED)>0      
    
    @property
    def Text(self):
        pmbuf = ProcessMem(processId=self._pid, buffer_size=win32con.MAX_PATH)
        pmbuf.write(ctypes.c_char_p(six.b('\0') * win32con.MAX_PATH), win32con.MAX_PATH)                       
        ptext = ctypes.c_char_p(six.b('\0') * win32con.MAX_PATH)
        tvi = wintypes.TVITEM()
        tvi.mask = commctrl.TVIF_TEXT
        tvi.pszText = pmbuf.Buffer
        tvi.cchTextMax = win32con.MAX_PATH
        tvi.hItem = self.__item
        tvisize = ctypes.sizeof(wintypes.TVITEM)        
        pmtvi = ProcessMem(processId=self._pid, buffer_size=tvisize)        
        pmtvi.write(ctypes.byref(tvi), tvisize)        
        win32gui.SendMessage(self.HWnd, commctrl.TVM_GETITEM, 0, pmtvi.Buffer)
        pmbuf.read(ptext, win32con.MAX_PATH)    
        return util.myEncode(ptext.value, 'utf-8', 'gbk') 
    
    @property
    def NextSibling(self):
        return TreeViewItem(self.HWnd, win32gui.SendMessage(self.HWnd, commctrl.TVM_GETNEXTITEM, commctrl.TVGN_NEXT, self.__item))


class _ITEMLIST(list):
    def __getitem__(self, key):
        if isinstance(key, int):
            cnt = len(self)
            if key >= cnt:
                raise IndexError("key(%d)已超出下标范围!" % key)
            else:
                return list.__getitem__(self, key)
        elif isinstance(key, six.string_types):
            if(isinstance(key, six.text_type)):
                key = key.encode('utf8')
            for item in self:
                if item.Text == key:
                    return item
            
class MenuItem(Control):
    '''菜单项控件。不要直接实例化这个类，而通过Menu.MenuItems来获取。
    '''
    class EnumMenuItemState(object):
        DISABLED = "Disabled"
        GRAYED = "Grayed"
        NORMAL = "Normal"
        UNKNOWN = "Unknown"
        
    def __init__(self, menu, index):
        MN_GETHMENU = 0x01E1
        self.menu = menu
        self.hMenu = win32gui.SendMessage(menu.HWnd, MN_GETHMENU, 0, 0)
        self.idx = index
        
    @property
    def IsSeperator(self):
        """返回该子项是否是分割线
        """
        uItemState = win32gui.GetMenuState(self.hMenu, self.idx, win32con.MF_BYPOSITION)
        return (uItemState & win32con.MF_SEPARATOR) > 0
    
    @property
    def State(self):
        """返回菜单项的状态
        
        :rtype: EnumMenuItemState
        """
        uItemState = win32gui.GetMenuState(self.hMenu, self.idx, win32con.MF_BYPOSITION)
        if (uItemState & win32con.MF_GRAYED) > 0:
            return MenuItem.EnumMenuItemState.GRAYED
        elif (uItemState & win32con.MF_DISABLED) > 0:
            return MenuItem.EnumMenuItemState.DISABLED
        elif uItemState == -1:
            return MenuItem.EnumMenuItemState.UNKNOWN
        else:
            return MenuItem.EnumMenuItemState.NORMAL
    
    @property
    def Text(self):
        """可见文本
        """
        if not self.IsSeperator and self.State == MenuItem.EnumMenuItemState.NORMAL:
            pText = ctypes.c_char_p(six.b('\0') * win32con.MAX_PATH)
            ctypes.windll.user32.GetMenuStringA(self.hMenu, 
                                                self.idx, 
                                                pText, 
                                                win32con.MAX_PATH, 
                                                win32con.MF_BYPOSITION)
            text = pText.value
            os_encoding = locale.getdefaultlocale(None)[1]
            try:
                if six.PY3:
                    return text.decode(os_encoding)
                return text.decode(os_encoding).encode('utf-8')
            except UnicodeDecodeError:
                return text
        return None
            
    @staticmethod
    def __enum_childwin_callback(hwnd, hwnds):
        if hwnd and win32gui.IsWindowVisible(hwnd):
            hwnds.append(hwnd)
    
    def __getSysMenuWindow(self):
        hwnds = []
        win32gui.EnumWindows(MenuItem.__enum_childwin_callback, hwnds)
        menuWnds = []
        for hwnd in hwnds:
            if '#32768' == Window(root=hwnd).ClassName:
                menuWnds.append(hwnd)
        return len(menuWnds)
    
    @property
    def SubMenu(self):
        """鼠标移动到该子项上，产生子菜单，并返回该子菜单
        
        :rtype: Menu
        """
        MN_GETHMENU = 0x01E1
        x, y = self.BoundingRect.Center.All
        menuWndsCnt = self.__getSysMenuWindow()
        Mouse.click(x, y)
        self._timeout.retry(self.__getSysMenuWindow, (), (), lambda x: x == menuWndsCnt+1)
        hSubMenuWnd = 0
        hSubMenu = win32gui.GetSubMenu(self.hMenu, self.idx)
        hwnds = []
        win32gui.EnumWindows(MenuItem.__enum_childwin_callback, hwnds)
        for hwnd in hwnds:
            hMenu = win32gui.SendMessage(hwnd, MN_GETHMENU, 0, 0)
            if hMenu == hSubMenu:
                hSubMenuWnd = hwnd
                break
        
        if hSubMenuWnd:
            return Menu(root=hSubMenuWnd)
    
    @property
    def BoundingRect(self):
        """返回rect
        """
        left, top, right, bottom = win32gui.GetMenuItemRect(self.menu.HWnd, self.hMenu, self.idx)[-1]
        return util.Rectangle((left, top, right, bottom))
        
    def click(self, mouseFlag=MouseFlag.LeftButton, 
                    clickType=MouseClickType.SingleClick,
                    xOffset=None,
                    yOffset=None):
        """点击控件
        
        :type mouseFlag: qt4c.mouse.MouseFlag
        :param mouseFlag: 鼠标按钮
        :type clickType: qt4c.mouse.MouseClickType 
        :param clickType: 鼠标动作
        :type xOffset: int
        :param xOffset: 距离控件区域左上角的偏移。
                                                               默认值为None，代表控件区域x轴上的中点；
                                                               如果为负值，代表距离控件区域右上角的x轴上的绝对值偏移；
        :type yOffset: int
        :param yOffset: 距离控件区域左上角的偏移。
                                                               默认值为None，代表控件区域y轴上的中点；
                                                              如果为负值，代表距离控件区域右上角的y轴上的绝对值偏移；
        """
        x, y = self._getClickXY(xOffset, yOffset)
        Mouse.click(x,y, mouseFlag, clickType)
        # Mouse.sendClick(self.menu.HWnd, x,y, mouseFlag, clickType)
            
class Menu(Window):
    '''
            菜单控件
    '''
    def __init__(self, root=None, locator=None):
        '''Constructor，当root和locator都是None时找到当前的唯一显示的菜单
        
        :param root: root控件
        :param locator: 定位参数 
        '''
        if root==None and locator==None:
            control.Control.__init__(self)
            hwnd = Menu._timeout.retry(self.__findSysMenuWindow,(),(),lambda x: x != None)
            root = Control(root=hwnd)
        
        Window.__init__(self, root, locator)
    
    def _getSubMenuItemsCount(self):
        MN_GETHMENU = 0x01E1
        hMenu = win32gui.SendMessage(self.HWnd, MN_GETHMENU, 0, 0)
        if hMenu:
            return win32gui.GetMenuItemCount(hMenu)
        return 0
    
    def __findSysMenuWindow(self):
#        import exceptions
        hwnds = []
        win32gui.EnumWindows(Menu.__enum_childwin_callback, hwnds)
        menuWnds = []
        for hwnd in hwnds:
            if '#32768' == Window(root=hwnd).ClassName:
                menuWnds.append(hwnd)
        if not menuWnds:
            return None
        menuCount = len(menuWnds)
        if menuCount > 1:
            raise ControlAmbiguousError("%d menu windows found!" % menuCount)
        return menuWnds[0]
    
    @staticmethod
    def closeAllSysMenuWindow():
        """关闭所有的系统menu窗口
        """
        hwnds = []
        win32gui.EnumWindows(Menu.__enum_childwin_callback, hwnds)
        for hwnd in hwnds:
            menuWnd = Window(root=hwnd)
            if '#32768' == menuWnd.ClassName:
                menuWnd.close()
        
    @staticmethod
    def __enum_childwin_callback(hwnd, hwnds):
        if hwnd and win32gui.IsWindowVisible(hwnd):
            hwnds.append(hwnd)
    
    def __iter__(self):
        cnt = self._getSubMenuItemsCount()
        for i in range(cnt):
            yield MenuItem(self, i)
            
    def __getitem__(self, key):
        '''根据key返回MenuItem
        key: 菜单索引或菜单文字
        '''
        self._timeout.retry(self._getSubMenuItemsCount, (), (), lambda x: x > 0)
        
        cnt = self._getSubMenuItemsCount()
        if isinstance(key, six.integer_types):
            if key >= cnt:
                raise IndexError("key(%d)超出下标范围!" % key)
            return MenuItem(self, key)
        if isinstance(key, six.string_types):
            for i in range(cnt):
                mi = MenuItem(self, i)
                if mi.Text == key:
                    return mi
            raise IndexError("item(%s) not found!" % key)
        raise TypeError("type(%s) error!" % str(type(key)))
       
    @property
    def MenuItems(self):
        '''获取MenuItem。通过MenuItems[菜单项索引]或MenuItems[菜单项文字]返回MenuItem实例。
        '''
        return self


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''
Web自动化公共接口
'''
import win32api
import win32con

try:
    from qt4w import XPath
    from qt4w.webcontrols import WebElement,FrameElement
    from browser import EnumBrowserType as BrowserEnum
    from qt4w.webcontrols import WebPage as BaseWebPage
    
    class WebPage(BaseWebPage):
        '''封装PC端自动化所需的页面相关的逻辑
        '''
        def __init__(self, webview):
            '''构造函数
            
            :type webview: webview
            :param webview: 包含webdriver的webview对象
            '''
            self.page = webview
            super(WebPage, self).__init__(webview)
            
        def close(self):
            win32api.PostMessage(self.page._window.HWnd,win32con.WM_CLOSE,0,0)
            
        def activate(self):
            self.page.activate()
except:
    raise RuntimeError('QT4W Not Found')



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QT4C available.  
# Copyright (C) 2020 THL A29 Limited, a Tencent company.  All rights reserved.
# QT4C is licensed under the BSD 3-Clause License, except for the third-party components listed below. 
# A copy of the BSD 3-Clause License is included in this file.
#
'''
使用UIA方式去访问控件
'''

import time
import six
from comtypes.client import CreateObject, GetModule
from ctypes import *

from testbase import logger
from testbase.util import LazyInit

from qt4c import wincontrols
from qt4c import control
from qt4c.util import Rectangle,Timeout
from qt4c.mouse import Mouse,MouseFlag,MouseClickType
from qt4c.keyboard import Keyboard
from qt4c.exceptions import ControlAmbiguousError, ControlNotFoundError, ControlExpiredError, TimeoutError


IUIAutomation = GetModule("UIAutomationCore.dll")
UIAutomationClient = CreateObject("{ff48dba4-60ef-4201-aa87-54103eef594e}", interface=IUIAutomation.IUIAutomation)
RawWalker = UIAutomationClient.RawViewWalker
ControlWalker = UIAutomationClient.controlViewWalker
UIAControlType={50000: 'Button',
                 50001: 'Calendar', 
                 50002: 'CheckBox', 
                 50003: 'ComboBox', 
                 50004: 'Edit', 
                 50005: 'Hyperlink', 
                 50006: 'Image', 
                 50007: 'ListItem', 
                 50008: 'List', 
                 50009: 'Menu', 
                 50010: 'MenuBar', 
                 50011: 'MenuItem', 
                 50012: 'ProgressBar', 
                 50013: 'RadioButton', 
                 50014: 'ScrollBar', 
                 50015: 'Slider', 
                 50016: 'Spinner', 
                 50017: 'UStatusBar', 
                 50018: 'Tab', 
                 50019: 'TabItem', 
                 50020: 'Text', 
                 50021: 'ToolBar', 
                 50022: 'ToolTip', 
                 50023: 'Tree', 
                 50024: 'TreeItem', 
                 50025: 'Custom control type', 
                 50026: 'Group', 
                 50027: 'Thumb', 
                 50028: 'DataGrid', 
                 50029: 'DataItem', 
                 50030: 'Document', 
                 50031: 'SplitButton', 
                 50032: 'Window', 
                 50033: 'Pane', 
                 50034: 'Header', 
                 50035: 'HeaderItem', 
                 50036: 'Table', 
                 50037: 'TitleBar', 
                 50038: 'Separator', 
                 50039: 'SemanticZoom', 
                 50040: 'AppBar'}

def find_UIAElm(Condition,timeout=10):
    start = time.time()
    try_count = 0
    while time.time() - start < timeout:
        try:
            desk_elm = UIAutomationClient.GetRootElement()
            elm = desk_elm.FindFirst(IUIAutomation.TreeScope_Descendants,Condition)
            elm.CurrentName#验证获取到是不是空elm
            return elm
        except ValueError:
            logger.info("未查找到有效uia元素，尝试重新查找")
            time.sleep(0.5)
            try_count += 1
    raise TimeoutError("在%d秒里尝试了%d次" %(timeout,try_count))



class Control(control.Control):
    '''
    UIA方式访问控件基类
    '''
    def __init__(self,root=None,locator=None):
        '''构造函数        
        :type root: UIA.Control or None
        :param root: 开始查找的UIA控件或包含UIA的win32control.Window；
        :type locator: str or qt4c.qpath.QPath(后面支持）
        :param locator: UIA控件的name属性  or qpath(后面支持）
        :attention: 参数root和locator不能同时为None
        '''
        control.Control.__init__(self)
    
        if locator is None and root is None:
            raise RuntimeError("传入参数locator和root不能同时为None!")
        self._root = root
        self._locator = locator
        self._uiaobj = LazyInit(self,'_uiaobj',self._init_uiaobj)
        
    def _init_uiaobj(self):
        '''初始化uia对象
        '''
        if self._root is None:
            self._root = Control(root=UIAutomationClient.GetRootElement())
            self._root.empty_invoke()
            
            
        if isinstance(self._root, wincontrols.Control):
                pid = self._root.ProcessId
                if not pid or not self._root.Valid:
                    raise ControlExpiredError("父控件/父窗口已经失效，查找中止！")
                if self._locator is None:
                    cnd = UIAutomationClient.CreatePropertyCondition(IUIAutomation.UIA_NativeWindowHandlePropertyId,self._root.HWnd)
                    self._root = find_UIAElm(cnd)
                            
        if self._locator is None:
            if isinstance(self._root, Control):
                self._root.empty_invoke()
                elm = self._root._uiaobj
            elif isinstance(self._root, IUIAutomation.IUIAutomationElement):
                elm = self._root
            else:
                raise TypeError("root应为uiacontrols.Control类型或者UIA element，实际类型为：%s" % type(self._root))
        else:                        
            if isinstance(self._locator, six.string_types):
                if isinstance(self._root, Control):
                    args = (self._root,self._locator)
                    try:
                        from qt4c.qpath import _find_by_name
                        foundctrl = self._timeout.retry(_find_by_name, args, (ControlNotFoundError))
                        foundctrl.empty_invoke()
                        elm = foundctrl._uiaobj
                    except TimeoutError:
                        raise  ControlNotFoundError("找不到name为 (%s)的UIA子控件！" % self._locator)
                else:
                    raise TypeError("root应为uiacontrols.Control类型，实际类型为：%s" % type(self._root))
            else:
                try:
                    kwargs = {'root':self._root}
                    foundctrls =  self._timeout.retry(self._locator.search, kwargs, (), lambda x: len(x)>0)
                except TimeoutError as erro:
                    raise ControlNotFoundError("<%s>中的%s查找超时：%s" % (self._root,self._locator.getErrorPath(),erro))
                nctrl = len(foundctrls)
                if(nctrl>1):
                    raise ControlAmbiguousError("<%s>找到%d个控件" % (self._locator, nctrl))
                foundctrls[0].empty_invoke()
                elm = foundctrls[0]._uiaobj
        return elm
    
    def empty_invoke(self):
        '''无意义调用，用于主动初始化对象
        '''
        self._uiaobj.CurrentName
        
    def SetFocus(self):
        self._uiaobj.SetFocus()#这个接口没有作用

    @property
    def Valid(self):
        """是否是有效控件
        """
        return self.Enabled
    
    @property
    def Width(self):
        """控件宽度
        """
        rect = self._getrect()
        return rect['Width']
        
    @property
    def Height(self):
        """控件高度
        """
        rect = self._getrect()
        return rect['Height']
    
    @property    
    def BoundingRect(self):
        """返回控件
        """
        rect = self._getrect()
        return Rectangle((int(rect['Left']),
                          int(rect['Top']),
                          int(rect['Left']+rect['Width']),
                          int(rect['Top']+rect['Height'])
                          ))
        
    def _getrect(self):
        rect = {'Left':0,'Top':0,'Width':0,'Height':0}
        ( rect['Left'], rect['Top'], rect['Width'], rect['Height'])=self._uiaobj.GetCurrentPropertyValue(IUIAutomation.UIA_BoundingRectanglePropertyId)
        return rect
    
    def click(self, mouseFlag = MouseFlag.LeftButton,
              clickType=MouseClickType.SingleClick,
              xOffset=None, yOffset=None):
        """点击控件
        :type mouseFlag:qt4c.mouse.MouseFlag
        :param mouseFlag: 鼠标按钮类型
        :type clickType:qt4c.mouse.MouseClickType
        :param clickType:点击动作类型
        :type xOffset:int
        :param 横向偏移量
        :type yOffset:int
        :param 纵向偏移量
        """
        Timeout(5,0.5).waitObjectProperty(self, 'Enabled', True)
        x, y = self._getClickXY(xOffset, yOffset)
        Mouse.click(int(x),int(y), mouseFlag, clickType)
        
    @property
    def ProcessId(self):
        pid = self._uiaobj.GetCurrentPropertyValue(IUIAutomation.UIA_ProcessIdPropertyId)
        return pid
        
    @property
    def ControlType(self):
        """返回UIA控件的类型
        """
        typeint = self._uiaobj.GetCurrentPropertyValue(IUIAutomation.UIA_ControlTypePropertyId)
        return UIAControlType[typeint]
    
    @property
    def Enabled(self):
        """此控件是否可用
        """
        return self._uiaobj.GetCurrentPropertyValue(IUIAutomation.UIA_IsEnabledPropertyId)
    
    @property
    def Children(self):
        """返回子控件列表
        :rtype ListType
        """
        self.empty_invoke()
        children = []
        child = RawWalker.GetFirstChildElement(self._uiaobj)
        while(child != None):
            try:
                child.CurrentName#测试uiaelm对象是否有效
                children.append(Control(root=child))
                child = RawWalker.GetNextSiblingElement(child)
            except ValueError:
                child= None
        return children
    
    @property
    def Parent(self):
        self.empty_invoke()
        parent = RawWalker.GetParentElement(self._uiaobj)
        try:
            parent.CurrentName#测试uia elm对象是否有效
        except ValueError:
            parent =None
        else:
            return Control(root=parent)
    
    @property
    def Name(self):
        """返回control的name属性
        """
        return self._uiaobj.CurrentName
    
    @property
    def Type(self):
        """返回控件类型
        """
        return self._uiaobj.CurrentLocalizedControlType
    
    @property
    def Value(self):
        """返回控件value属性（通常是文本信息）
        """
        return self._uiaobj.GetCurrentPropertyValue(IUIAutomation.UIA_LegacyIAccessibleValuePropertyId)
    
    def equal(self, other):
        if not isinstance(other, Control):
            return False
        if self.ProcessId != other.ProcessId:
            return False
        return self._uiaobj == other._uiaobj
    
    def exist(self):
        """判断控件是否存在
        """
        try:
            self._init_uiaobj()
        except Exception as e:
            print("判断UIA控件存在失败,UIA elm实例化异常:%s" %e)
            return False
        return True
    
    def wait_for_exist(self,timeout,interval):
        """等待控件存在
        """
        Timeout(timeout,interval).retry(self.exist,(),(),lambda x:x==True)
    
    @property    
    def Hwnd(self):
        """返回控件句柄
        """
        return self.hwnd

    @property
    def HWnd(self):
        """兼容其他类型控件
        """
        return self.hwnd

    @property
    def hwnd(self):
        return self._uiaobj.CurrentNativeWindowHandle
    
    @property
    def HasKeyboardFocus(self):
        """返回是否被选为键盘输入
        """
        return self._uiaobj.CurrentHasKeyboardFocus
    
    @property
    def ClassName(self):
        """返回ClassName
        """
        return self._uiaobj.CurrentClassName   
   

        
        
            
class UIAWindows(Control, control.ControlContainer):
    '''UIA控件窗体定义
    '''
    def __init__(self,root=None, locator=None):
        Control.__init__(self,root=root,locator=locator)
        control.ControlContainer.__init__(self)
    
    
    @property
    def Window(self):
        """返回wincontrols.Window
        """   
        return wincontrols.Window(root=self.Hwnd)

    @property
    def Visible(self):
        """窗口是否可见
        """
        if not self.Valid:
            return False
        return self.Window.Visible
    
    @property
    def Minimized(self):
        """该窗口是否最小化
        """
        if not self.Valid:
            raise RuntimeError("句柄无效，窗口已销毁!")
        return self.Window.Minimized
    
    


class Edit(Control):
    
    def input(self,data):
        """对Edit类型控件进行输入
        :type keys: utf-8 str or unicode
        :param keys: 键盘输入字符串,可输入组合键，如"{CTRL}{MENU}a"
        """
        self.click()
        time.sleep(0.5)
        keys = "{CTRL}a"
        Keyboard.inputKeys(keys)
        keys = "{DEL}"
        Keyboard.inputKeys(keys)
        Keyboard.inputKeys(data)

class ComboBox(Control):
    
    def input(self,data):
        """对Edit类型控件进行输入
        :type keys: utf-8 str or unicode
        :param keys: 键盘输入字符串,可输入组合键，如"{CTRL}{MENU}a"
        """
        self.click()
        time.sleep(0.5)
        keys = "{CTRL}a"
        Keyboard.inputKeys(keys)
        keys = "{DEL}"
        Keyboard.inputKeys(keys)
        Keyboard.inputKeys(keys)
        Keyboard.inputKeys(data)

    def expand(self):
        pattern = self._uiaobj.GetCurrentPattern(IUIAutomation.UIA_ExpandCollapsePatternId)
        dllpoint = cast(pattern, POINTER(IUIAutomation.IUIAutomationExpandCollapsePattern))
        ret = dllpoint.Expand() == 0
        return ret

    def collapse(self):
        pattern = self._uiaobj.GetCurrentPattern(IUIAutomation.UIA_ExpandCollapsePatternId)
        dllpoint = cast(pattern, POINTER(IUIAutomation.IUIAutomationExpandCollapsePattern))
        ret = dllpoint.Collapse() == 0
        return ret

    @property
    def Value(self):
        return self._uiaobj.GetCurrentPropertyValue(IUIAutomation.UIA_LegacyIAccessibleValuePropertyId)

    @Value.setter
    def Value(self, _item):
        """对Combobox类型控件设置value值
        :type keys: utf-8 str or unicode
        :param keys: 被设置选项的值
        """
        pattern = self._uiaobj.GetCurrentPattern(IUIAutomation.UIA_ValuePatternId)
        dllpoint = cast(pattern, POINTER(IUIAutomation.IUIAutomationValuePattern))
        ret = dllpoint.SetValue(_item) == 0
        return ret

class RadioButton(Control):
    def is_select(self):
        pattern = self._uiaobj.GetCurrentPattern(IUIAutomation.UIA_SelectionItemPatternId)
        dllpoint = cast(pattern, POINTER(IUIAutomation.IUIAutomationSelectionItemPattern))
        return dllpoint.CurrentIsSelected
        
if __name__ == '__main__':
    pass
